from sqlalchemy import Column, String, UUID
import uuid as uuid_type
from joule.models.meta import Base


class Node(Base):
    """
    Attributes:
        id (uuid): globally unique identifier, not visible to user
        private_key (str): --not used yet--
        public_key (str): --not used yet--
    """
    __tablename__ = 'node'
    __table_args__ = {"schema": "metadata"}

    uuid: uuid_type.UUID = Column("uuid", UUID(as_uuid=True), primary_key=True, default=uuid_type.uuid4, unique=True)
    private_key: str = Column(String)
    public_key: str = Column(String)
    version: str = Column(String)

    


import unittest

from joule import api, errors


class TestModuleMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()

    async def asyncTearDown(self) -> None:
        await self.node.close()

    async def test_module_get(self):
        module = await self.node.module_get("plus1")
        output_path = module.outputs['output']
        input_path = module.inputs['input']
        self.assertEqual(output_path, '/live/plus1')
        self.assertEqual(input_path, '/live/base')

    async def test_module_get_errors(self):
        with self.assertRaises(errors.ApiError):
            await self.node.module_get("bad")

    async def test_module_list(self):
        modules = await self.node.module_list()
        self.assertEqual(len(modules),2)
        for m in modules:
            self.assertIs(type(m), api.Module)

    async def test_module_logs(self):
        logs = await self.node.module_logs("plus1")
        self.assertIn("starting", logs[0])

    async def test_module_logs_errors(self):
        with self.assertRaises(errors.ApiError):
            await self.node.module_logs("bad")



from aiohttp import web
from joule.utilities import ConnectionInfo
from joule.models.supervisor import Supervisor
from joule.models import DataStore, EventStore
from sqlalchemy.orm import Session
import uuid as uuid_type
module_connection_info = web.AppKey("module-connection-info", ConnectionInfo)
supervisor = web.AppKey("supervisor", Supervisor)
data_store = web.AppKey("data-store", DataStore)
event_store = web.AppKey("event-store", EventStore)
db = web.AppKey("db", Session)
base_uri = web.AppKey("base-uri", str)
name = web.AppKey("name", str)
port = web.AppKey("port", int)
scheme = web.AppKey("scheme", str)
cafile = web.AppKey("cafile", str)
remote_ip = web.AppKey("remote-ip", str)  # set by middleware but not used?
# globally unique ID for a node, automatically generated
# used to identify where data originated when copying between nodes
uuid = web.AppKey("uuid", uuid_type.UUID) 




from joule.models import module, DataStream, Element
from joule.errors import ConfigurationError
from tests import helpers
import unittest


class TestModule(unittest.TestCase):

    def setUp(self):
        self.config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = long text
                 exec_cmd = /bin/runit.sh
                 is_app = yes
               [Arguments]
                 arg1 = val1
                 arg2 = val2
            """)

# Test module parsing (from_config)

    def test_parses_base_config(self):
        m = module.from_config(self.config)
        self.assertEqual(m.name, "test")
        self.assertEqual(m.description, "long text")
        self.assertEqual(m.exec_cmd, "/bin/runit.sh")
        self.assertTrue(m.is_app)
        self.assertEqual(len(m.arguments), 2)  # default value
        self.assertEqual(m.arguments["arg1"], "val1")
        self.assertEqual(m.arguments["arg2"], "val2")

    def test_uses_default_values(self):
        self.config.remove_section("Arguments")
        self.config.remove_option("Main", "is_app")
        self.config.remove_option("Main", "description")
        m = module.from_config(self.config)
        self.assertEqual(m.description, "")
        self.assertFalse(m.is_app)
        self.assertEqual(len(m.arguments), 0)

    def test_requires_name_setting(self):
        self.config.remove_option("Main", "name")
        with self.assertRaisesRegex(ConfigurationError, "name"):
            module.from_config(self.config)

    def test_has_json_representation(self):
        # create an input stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32)
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # create an output stream
        dest = DataStream(id=1, name="dest", keep_us=100, datatype=DataStream.DATATYPE.INT16)
        dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.EVENT) for x in range(5)]
        m = module.from_config(self.config)
        m.inputs = {'input': src}
        m.outputs = dict(output=dest)
        result = m.to_json()
        # make sure basic attributes are in the output
        self.assertEqual(result['name'],'test')
        self.assertEqual(result['exec_cmd'],'/bin/runit.sh')
        # make sure inputs are included (name: id)
        self.assertEqual(result['inputs']['input'],0)
        self.assertEqual(result['outputs']['output'],1)



import unittest
import configparser
import tempfile
import asyncio

from joule.services import load_config
from joule.errors import ConfigurationError


class TestLoadConfigErrors(unittest.TestCase):

    def test_error_if_directories_do_not_exist(self):
        with tempfile.TemporaryDirectory() as good_dir1:
            with tempfile.TemporaryDirectory() as good_dir2:
                bad_dir = "/does/not/exist"
                bad_module_dir = (bad_dir, good_dir1, good_dir2)
                bad_stream_dir = (good_dir1, bad_dir, good_dir2)
                for setup in zip(bad_module_dir, bad_stream_dir):
                    parser = configparser.ConfigParser()
                    parser.read_string("""
                                [Main]
                                ModuleDirectory=%s
                                StreamDirectory=%s
                            """ % setup)
                    with self.assertRaises(ConfigurationError):
                        load_config.run(custom_values=parser)

    def test_error_if_missing_database_configuration(self):
        with tempfile.TemporaryDirectory() as module_dir:
            with tempfile.TemporaryDirectory() as stream_dir:
                with tempfile.TemporaryDirectory() as sock_dir:
                    parser = configparser.ConfigParser()
                    parser.read_string("""
                                [Main]
                                ModuleDirectory=%s
                                StreamDirectory=%s
                                SocketDirectory=%s
                            """ % (module_dir, stream_dir, sock_dir))
                    with self.assertRaisesRegex(ConfigurationError, "Database"):
                        load_config.run(custom_values=parser)

    def test_error_on_bad_ip_address(self):
        parser = configparser.ConfigParser()
        bad_ips = ["8.8.x.8", "bad", "", "900.8.4.100"]
        for ip in bad_ips:
            parser.read_string("""
                [Main]
                IPAddress=%s
            """ % ip)
            with self.assertRaisesRegex(ConfigurationError, "IPAddress"):
                load_config.run(custom_values=parser, verify=False)

    def test_error_on_bad_port(self):
        parser = configparser.ConfigParser()
        bad_ports = ["-3", "99999", "abc", ""]
        for port in bad_ports:
            parser.read_string("""
                [Main]
                IpAddress=127.0.0.1
                Port=%s
            """ % port)
            with self.assertRaisesRegex(ConfigurationError, "Port"):
                load_config.run(custom_values=parser, verify=False)

    def test_errors_on_invalid_insert_period(self):
        bad_periods = ['-1', '0', 'abc', '']
        for period in bad_periods:
            parser = configparser.ConfigParser()
            parser.read_string("""
                        [Main]
                        InsertPeriod = %s
                        """ % period)
            with self.assertRaisesRegex(ConfigurationError, "InsertPeriod"):
                load_config.run(custom_values=parser, verify=False)

    def test_errors_on_invalid_cleanup_period(self):
        bad_periods = ['-1', '0', 'abc', '', '30']
        for period in bad_periods:
            parser = configparser.ConfigParser()
            parser.read_string("""
                        [Main]
                        InsertPeriod = 50
                        CleanupPeriod = %s
                        """ % period)
            with self.assertRaisesRegex(ConfigurationError, "CleanupPeriod"):
                load_config.run(custom_values=parser, verify=False)

    def test_errors_on_invalid_max_log_lines(self):
        parser = configparser.ConfigParser()
        for bad_val in [-1, 'abc']:
            parser.read_string("""
                                    [Main]
                                    MaxLogLines = %s
                                    """ % bad_val)
            with self.assertRaisesRegex(ConfigurationError, "MaxLogLines"):
                load_config.run(custom_values=parser, verify=False)





from aiohttp.test_utils import AioHTTPTestCase
from aiohttp import web
import aiohttp
import numpy as np
from sqlalchemy.orm import Session
import asyncio
from joule.models import DataStream, pipes
import joule.controllers
from tests.controllers.helpers import create_db, MockStore, MockSupervisor
from tests import helpers
from joule import app_keys
from joule.constants import EndPoints
from joule.errors import EmptyPipeError

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestDataController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app[app_keys.db], app[psql_key] = create_db(["/folder1/stream1:float32[x, y, z]",
                                            "/folder2/deeper/stream2:int16[val1, val2]"])
        app[app_keys.data_store] = MockStore()
        self.supervisor = MockSupervisor()
        app[app_keys.supervisor] = self.supervisor
        return app


    async def test_read_binary_data(self):
        db: Session = self.app[app_keys.db]
        store: MockStore = self.app[app_keys.data_store]
        nchunks = 3
        nintervals=4
        store.configure_extract(nchunks, nintervals)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={"path": "/folder1/stream1"})
        rx_chunks = 0
        # this data is binary so we can't do much with it, just make sure 
        # the handler is called multiple times :) 
        async for data, _ in resp.content.iter_chunks():
            if len(data) > 0:
                rx_chunks += 1
        self.assertGreater(rx_chunks, 1)

        # can retrieve stream by id and as decimated data
        stream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={"id": stream.id, 'decimation-level': 16})
        rx_chunks = 0
        async for data, _ in resp.content.iter_chunks():
            if len(data) > 0:
                rx_chunks += 1
        self.assertGreater(rx_chunks, 1)
        self.assertEqual(resp.headers['joule-decimation'], '16')
        self.assertEqual(resp.headers['joule-layout'], stream.decimated_layout)


    async def test_read_json_data(self):
        db: Session = self.app[app_keys.db]
        store: MockStore = self.app[app_keys.data_store]
        n_chunks = 10
        n_intervals = 5
        # mock data has 5 intervals retrieved as 10 chunks each
        store.configure_extract(n_chunks, n_intervals)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_json, params={"path": "/folder1/stream1"})
        data = await resp.json()
        self.assertEqual(1, data['decimation_factor'])
        self.assertEqual(len(data['data'][0]), n_chunks * 25)
        # can retrieve stream by id and as decimated data as well
        stream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_json, params={"id": stream.id, 'decimation-level': 16})
        data = await resp.json()
        self.assertEqual(data['decimation_factor'], 16)
        # only the interval is exposed, the chunking is transparent
        self.assertEqual(len(data['data']), n_intervals)
        for interval in data['data']:
            self.assertEqual(len(interval), n_chunks * 25)


    async def test_subscribes_to_data(self):
        db: Session = self.app[app_keys.db]
        my_stream = db.query(DataStream).filter_by(name="stream1").one()
        blk1 = helpers.create_data(my_stream.layout)
        blk2 = helpers.create_data(my_stream.layout, length=50)
        my_pipe = pipes.LocalPipe(my_stream.layout)
        my_pipe.write_nowait(blk1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(blk2)
        my_pipe.close_nowait()
        self.supervisor.subscription_pipe = my_pipe
        async with self.client.get(EndPoints.data, params={"id": my_stream.id,
                                                    "subscribe": '1'}) as resp:
            pipe = pipes.InputPipe(stream=my_stream, reader=resp.content)
            rx_blk1 = await pipe.read()
            pipe.consume(len(rx_blk1))
            np.testing.assert_array_equal(blk1, rx_blk1)
            self.assertTrue(pipe.end_of_interval)

            rx_blk2 = await pipe.read()
            pipe.consume(len(rx_blk2))
            np.testing.assert_array_equal(blk2, rx_blk2)
            with self.assertRaises(EmptyPipeError):
                await pipe.read()
        self.assertEqual(self.supervisor.unsubscribe_calls, 1)


    async def test_unsubscribes_terminated_connections(self):
        db: Session = self.app[app_keys.db]
        supervisor: MockSupervisor = self.app[app_keys.supervisor]
        supervisor.hang_pipe = True
        my_stream = db.query(DataStream).filter_by(name="stream1").one()
        my_pipe = pipes.LocalPipe(my_stream.layout)
        my_pipe.close_nowait()
        self.supervisor.subscription_pipe = my_pipe
        async with self.client.get(EndPoints.data, params={"id": my_stream.id,
                                                    "subscribe": '1'}):
            # ignore the data
            pass
        # NOTE: unsubscribe is called but the exception propogates and somehow
        # we loose the reference so this assert fails. Not a problem but why??
        # self.assertEqual(self.supervisor.unsubscribe_calls, 1)


    async def test_write_data(self):
        db: Session = self.app[app_keys.db]
        store: MockStore = self.app[app_keys.data_store]
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        data = helpers.create_data(stream.layout)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={"path": "/folder1/stream1"},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.inserted_data)

        # can write stream by id as well
        store.inserted_data = False
        data = helpers.create_data(stream.layout)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={"id": stream.id},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.inserted_data)

        # can set merge_gap value
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={"id": stream.id, "merge-gap": 100},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertEqual(store.merge_gap, 100)

        # merge gap is 0 by default
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={"id": stream.id},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertEqual(store.merge_gap, 0)



    async def test_remove_data(self):
        db: Session = self.app[app_keys.db]
        store: MockStore = self.app[app_keys.data_store]
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data, params={"path": "/folder1/stream1",
                                                "start": "100", "end": "200"})
        self.assertEqual(resp.status, 200)
        (start, end) = store.removed_data_bounds
        self.assertEqual(start, 100)
        self.assertEqual(end, 200)

        # can remove data by path as well
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data, params={"id": stream.id,
                                                "start": "800", "end": "900"})
        self.assertEqual(resp.status, 200)
        (start, end) = store.removed_data_bounds
        self.assertEqual(start, 800)
        self.assertEqual(end, 900)


    async def test_interval_data(self):
        db: Session = self.app[app_keys.db]
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        resp = await self.client.get(EndPoints.data_intervals,
                                     params={"id": stream.id})
        # response should be a list of intervals
        data = await resp.json()
        self.assertGreater(len(data), 0)
        for interval in data:
            self.assertEqual(2, len(interval))

        # can query by path as well
        resp = await self.client.get(EndPoints.data_intervals,
                                     params={"path": "/folder1/stream1"})
        # response should be a list of intervals
        data = await resp.json()
        self.assertGreater(len(data), 0)
        for interval in data:
            self.assertEqual(2, len(interval))

        # can specify time bounds
        resp = await self.client.get(EndPoints.data_intervals,
                                     params={"path": "/folder1/stream1",
                                             "start": 10,
                                             "end": 20})
        # the mock store returns the start end bounds as the single interval
        data = await resp.json()
        self.assertEqual(data, [[10, 20]])

    async def test_consolidate_data(self):
        store: MockStore = self.app[app_keys.data_store]
        resp = await self.client.post(EndPoints.data_consolidate, params={"id": 1, "end": 200, "max_gap":100})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['num_consolidated'], 0)

        # check to make sure the store was called with the correct parameters
        self.assertTrue(store.consolidate_called)
        stream, start, end, max_gap = store.consolidate_params
        self.assertEqual(stream.id, 1)
        self.assertIsNone(start, None)
        self.assertEqual(end, 200)
        self.assertEqual(max_gap, 100)


    async def test_decimate_data(self):
        store: MockStore = self.app[app_keys.data_store]
        resp = await self.client.post(EndPoints.data_decimate, params={"id": 1})
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.redecimated_data)


    async def test_remove_decimations(self):
        store: MockStore = self.app[app_keys.data_store]
        resp = await self.client.delete(EndPoints.data_decimate, params={"id": 1})
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.dropped_decimations)



import aiohttp
import ssl
from aiohttp.client_exceptions import ClientError
from typing import Optional


# parser for boolean args
def yesno(val: str):
    """
    Convert a "yes" or "no" argument into a boolean value. Returns ``true``
    if val is "yes" and ``false`` if val is "no". Raises ValueError otherwise.
    This is function can be used as the **type** parameter for to handle module arguments
    that are "yes|no" flags.
    """
    if val is None:
        raise ValueError("must be 'yes' or 'no'")
    # standardize the string
    val = val.lower().strip()
    if val == "yes":
        return True
    elif val == "no":
        return False
    else:
        raise ValueError("must be 'yes' or 'no'")


async def detect_url(host, port: Optional[int] = None):  
    if port is not None:
        host = host + ":" + str(port)

    async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(5)) as session:
        # try to connect over https
        try:
            await session.get("https://" + host)
            return "https://" + host
        except ClientError:
            # try again over http
            try:
                await session.get("http://" + host)
                return "http://" + host
            # unable to determine the URL
            except ClientError:
                return None

def timestamps_are_monotonic(data, last_ts: Optional[int], name: str):
    import numpy as np
    if len(data) == 0:
        return True
    # if there are multiple rows, check that all timestamps are increasing
    if len(data) > 1 and np.min(np.diff(data['timestamp'])) <= 0:
        min_idx = np.argmin(np.diff(data['timestamp']))
        msg = ("Non-monotonic timestamp in new data to stream [%s] (%d<=%d)" %
               (name, data['timestamp'][min_idx + 1], data['timestamp'][min_idx]))
        print(msg)
        return False
    # check to make sure the first timestamp is larger than the previous block
    if last_ts is not None and (last_ts >= data['timestamp'][0]):
            msg = ("Non-monotonic timestamp between writes to stream [%s] (%d<=%d)" %
                   (name, data['timestamp'][0], last_ts))
            print(msg)
            return False
    return True

def validate_values(data):
    import numpy as np
    if np.isnan(data['timestamp']).any():
        return False
    if np.isnan(data['data']).any():
        return False
    return True



from aiohttp import web
import os
import io
from contextlib import redirect_stdout
import multiprocessing
import numpy as np
from typing import Dict, Optional, List
from aiohttp.test_utils import unused_port
import time
import signal
import asyncio
import json
import psutil
import tempfile


from tests import helpers
from joule.models import DataStream, StreamInfo, pipes, data_stream, master
from joule.api import annotation
from joule.constants import EndPoints, ConfigFiles
from joule.errors import EmptyPipeError
# from https://github.com/aio-libs/aiohttp/blob/master/examples/fake_server.py


class MockDbEntry:
    def __init__(self, stream: DataStream, info: StreamInfo, data: Optional[np.ndarray] = None,
                 intervals=None):
        self.stream = stream
        self.info = info
        self.data = data
        if intervals is None:
            intervals = []
        self.intervals = intervals

    def add_data(self, chunk):
        if self.data is None:
            self.data = chunk
        else:
            self.data = np.hstack((self.data, chunk))


class FakeJoule:

    def __init__(self):
        self.runner = None
        self.app = web.Application(middlewares=[self._authorize])

        self.app.router.add_routes(
            [
                web.get(EndPoints.root, self.info),
                # GET db_connection
                web.get(EndPoints.db_info, self.dbinfo),
                # GET version
                web.get(EndPoints.version_json, self.stub_get),
                # --- event stream routes ---
                web.get(EndPoints.event, self.stub_get),
                web.put(EndPoints.event_move, self.move_stream),
                web.put(EndPoints.event, self.update_stream),
                #web.post(EndPoints.event, None),
                web.delete(EndPoints.event, self.delete_stream),
                # --- event stream data routes ---
                #web.get(EndPoints.event_data, None),
                #web.get(EndPoints.event_data_count, None),
                #web.post(EndPoints.event_data, None),
                #web.delete(EndPoints.event_data, None),
                # --- stream routes ---

                web.get(EndPoints.stream, self.stream_info),
                web.put(EndPoints.stream_move, self.move_stream),
                web.put(EndPoints.stream, self.update_stream),
                web.post(EndPoints.stream, self.create_stream),
                web.delete(EndPoints.stream, self.delete_stream),
                # --- folder routes ---
                web.get(EndPoints.folders, self.stub_get),
                web.get(EndPoints.folder, self.folder_info),
                web.put(EndPoints.folder_move, self.move_folder),
                web.put(EndPoints.folder, self.update_folder),
                web.delete(EndPoints.folder, self.delete_folder),
                # --- data routes ---
                web.get(EndPoints.data, self.data_read),
                # GET data.json
                web.get(EndPoints.data_intervals, self.data_intervals),
                web.post(EndPoints.data_decimate, self.data_decimate),
                # DELETE data_decimate
                web.post(EndPoints.data_consolidate, self.data_consolidate),
                web.post(EndPoints.data, self.data_write),
                web.delete(EndPoints.data, self.data_delete),
                # --- module routes ---
                web.get(EndPoints.modules, self.stub_get),
                web.get(EndPoints.module, self.stub_get),
                web.get(EndPoints.module_logs, self.stub_get),
                # --- app routes ---
                # GET app_auth
                # GET app_json
                # --- proxy routes ---
                web.get(EndPoints.proxies, self.stub_get),
                web.get(EndPoints.proxy, self.stub_get),
                # -- master routes --
                web.get(EndPoints.masters, self.stub_get),
                web.post(EndPoints.master, self.create_master),
                web.delete(EndPoints.master, self.delete_master),
                # --- follower routes ---
                web.get(EndPoints.followers, self.stub_get),
                # POST follower
                web.delete(EndPoints.follower, self.stub_get),
                # -- annotation routes --
                web.get(EndPoints.annotations_info, self.get_annotation_info),
                web.get(EndPoints.annotations, self.get_annotations),
                web.put(EndPoints.annotation, self.update_annotation),
                web.post(EndPoints.annotation, self.create_annotation),
                web.delete(EndPoints.annotation, self.delete_annotation),
                web.delete(EndPoints.stream_annotations, self.delete_all_annotations),

                
                # --- event stream data routes ---

            ])
        self.stub_stream_info = False
        self.stub_stream_move = False
        self.stub_data_delete = False
        self.stub_stream_destroy = False
        self.stub_stream_create = False
        self.stub_data_read = False
        self.stub_data_write = False
        self.stub_folder_move = False
        self.stub_folder_destroy = False
        self.stub_folder_info = False
        self.stub_folder_update = False
        self.stub_data_intervals = False
        self.stub_data_consolidate = False
        self.stub_data_decimate = False
        self.stub_master = False
        self.stub_stream_update = False
        self.stub_annotation = False
        self.first_lumen_user = True
        self.response = ""
        # need this because node info cmd makes two requests
        self.dbinfo_response = "--fill-in--"
        self.num_consolidated = 5
        self.http_code = 200
        self.streams: Dict[str, MockDbEntry] = {}
        self.msgs = None
        self.key = "invalid"

    def start(self, port, msgs: multiprocessing.Queue, inbound_msgs, key):
        self.msgs = msgs
        self.key = key
        #f = io.StringIO()
        #with redirect_stdout(f):
        #web.run_app(self.app, host='127.0.0.1', port=port,
        #                handle_signals=True)
        #''
        #asyncio.run(self._run_server(port, inbound_msgs))
        asyncio.run(self._run_server(port, inbound_msgs))
        msgs.close()
        inbound_msgs.close()
        msgs.join_thread()
        inbound_msgs.join_thread()

    @web.middleware
    async def _authorize(self, request, handler):
        if 'X-API-KEY' not in request.headers:
            raise web.HTTPForbidden()
        key = request.headers['X-API-KEY']
        if key != self.key:
            raise web.HTTPForbidden()
        return await handler(request)

    async def _run_server(self, port, msgs):
        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', port)
        await site.start()

        while 1:
            await asyncio.sleep(0.1)
            if not msgs.empty():
                msgs.get()
                break
        await runner.cleanup()

    def add_stream(self, path, my_stream: DataStream, info: StreamInfo, data: Optional[np.ndarray],
                   intervals: Optional[List] = None):
        self.streams[path] = MockDbEntry(my_stream, info, data, intervals)

    async def create_stream(self, request: web.Request):
        if self.stub_stream_create:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        path = body['dest_path']
        if path == '':  # check for invalid value (eg)
            return web.Response(text='invalid request', status=400)
        new_stream = data_stream.from_json(body['stream'])
        if new_stream.id is None:
            new_stream.id = 150
        else:
            new_stream.id += 100  # give the stream  a unique id
        self.streams[path + '/' + new_stream.name] = MockDbEntry(new_stream, StreamInfo(None, None, None))
        return web.json_response(data=new_stream.to_json())

    async def update_stream(self, request: web.Request):
        if self.stub_stream_update:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        # so we can check that the message was received
        self.msgs.put(dict(body['stream']))
        return web.json_response(data={"stub": "value"})

    async def delete_stream(self, request: web.Request):
        if self.stub_stream_destroy:
            return web.Response(text=self.response, status=self.http_code)
        self.msgs.put(request.query['path'])
        return web.Response(text="ok")

    async def delete_folder(self, request: web.Request):
        if self.stub_folder_destroy:
            return web.Response(text=self.response, status=self.http_code)
        self.msgs.put({"path": request.query['path'],
                       "recursive": 'recursive' in request.query})
        return web.Response(text="ok")

    async def info(self, request: web.Request):
        return web.Response(text="Joule server")

    async def dbinfo(self, request: web.Request):
        return web.Response(text=self.dbinfo_response, status=self.http_code,
                            content_type='application/json')

    async def stub_get(self, request: web.Request):
        return web.Response(text=self.response, status=self.http_code,
                            content_type='application/json')

    async def stream_info(self, request: web.Request):
        if self.stub_stream_info or \
                ('path' in request.query and request.query['path'] == '/stub/response'):
            return web.Response(text=self.response, status=self.http_code,
                                content_type='application/json')
        try:
            mock_entry = self._find_entry(request.query)
        except ValueError:
            return web.Response(text="stream does not exist", status=404)
        stream_json = mock_entry.stream.to_json()
        stream_json['data_info'] = mock_entry.info.to_json()
        return web.json_response(stream_json)

    async def folder_info(self, request: web.Request):
        if self.stub_folder_info:
            return web.Response(text=self.response, status=self.http_code,
                                content_type='application/json')
        # this endpoint must be stubbed
        return web.Response(text="NOT IMPLEMENTED", status=500)

    async def data_delete(self, request: web.Request):
        if self.stub_data_delete:
            return web.Response(text=self.response, status=self.http_code)
        tag = '??'
        if 'path' in request.query:
            tag = request.query['path']
        elif 'id' in request.query:
            tag = request.query['id']
        start = None
        if 'start'in request.query:
            start = int(request.query['start'])
        end = None
        if 'end' in request.query:
            end = int(request.query['end'])
        self.msgs.put((tag, start, end))
        return web.Response(text="ok")

    async def data_read(self, request: web.Request):
        if self.stub_data_read:
            return web.Response(text=self.response, status=self.http_code)
        mock_entry = self._find_entry(request.query)
        if 'decimation-level' in request.query:
            layout = mock_entry.stream.decimated_layout
            decimation_level = request.query['decimation-level']
        elif 'max-rows' in request.query:
            layout = mock_entry.stream.decimated_layout
            decimation_level = 16 # just made up
        else:
            layout = mock_entry.stream.layout
            decimation_level = 1
        resp = web.StreamResponse(status=200,
                                  headers={'joule-layout': layout,
                                           'joule-decimation': str(decimation_level)})
        resp.enable_chunked_encoding()
        await resp.prepare(request)
        await resp.write(mock_entry.data.tobytes())
        return resp

    async def data_write(self, request: web.Request):
        if self.stub_data_write:
            return web.Response(text=self.response, status=self.http_code)

        if 'id' in request.query:
            stream_id = int(request.query['id'])
            mock_entry = [x for x in self.streams.values() if x.stream.id == stream_id][0]
        else:
            mock_entry = self.streams[request.query['path']]
        pipe = pipes.InputPipe(name="inbound", stream=mock_entry.stream, reader=request.content)
        istart = None
        iend = None
        while True:
            try:
                chunk = await pipe.read()
                if len(chunk) > 0:
                    if istart is None:
                        istart = chunk['timestamp'][0]
                    iend = chunk['timestamp'][-1]
                pipe.consume(len(chunk))
                if pipe.end_of_interval and istart is not None and iend is not None:
                    mock_entry.intervals.append([istart, iend])
                    istart = None
                    iend = None
                mock_entry.add_data(chunk)
            except EmptyPipeError:
                break
        self.msgs.put(mock_entry)
        return web.Response(text="ok")

    async def data_intervals(self, request: web.Request):
        if self.stub_data_intervals:
            return web.Response(text=self.response, status=self.http_code)
        entry = self._find_entry(request.query)
        intervals = [[int(x) for x in interval] for interval in entry.intervals]

        return web.json_response(data=intervals)

    async def data_consolidate(self, request: web.Request):
        if self.stub_data_consolidate:
            return web.Response(text=self.response, status=self.http_code)
        path = request.query['path']
        start = None
        if 'start'in request.query:
            start = int(request.query['start'])
        end = None
        if 'end' in request.query:
            end = int(request.query['end'])
        max_gap = int(request.query['max_gap'])
        self.msgs.put((path, start, end, max_gap))
        return web.json_response(data={"num_consolidated": self.num_consolidated})
    
    async def data_decimate(self, request: web.Request):
        if self.stub_data_decimate:
            return web.Response(text=self.response, status=self.http_code)
        path = request.query['path']
        self.msgs.put(('decimate',path))
        return web.Response(text="ok")
    
    async def move_stream(self, request: web.Request):
        if self.stub_stream_move:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        path = body['src_path']
        destination = body['dest_path']
        # so we can check that the message was received
        self.msgs.put((path, destination))
        return web.json_response(data={"stub": "value"})

    async def move_folder(self, request: web.Request):
        if self.stub_folder_move:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        path = body['src_path']
        destination = body['dest_path']
        # so we can check that the message was received
        self.msgs.put((path, destination))
        return web.json_response(data={"stub": "value"})

    async def update_folder(self, request: web.Request):
        if self.stub_folder_update:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        # so we can check that the message was received
        self.msgs.put(dict(body['folder']))
        return web.json_response(data={"stub": "value"})

    async def create_master(self, request: web.Request):
        if self.stub_master:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        if body["master_type"] == "user":
            self.msgs.put(body)
            return web.json_response({"key": "fakekey", "name": body["identifier"]})
        if body["master_type"] == "joule":
            self.msgs.put(body)
            return web.json_response({"name": body["identifier"]})
        if body["master_type"] == "lumen":
            if len(body["lumen_params"]) == 0:
                if self.first_lumen_user:
                    self.first_lumen_user = False
                    return web.Response(text="first_name")
                else:
                    return web.Response(text="auth_key", status=400)
            lumen_params = body["lumen_params"]
            self.msgs.put(lumen_params)
            return web.json_response({"name": body["identifier"]})

    async def delete_master(self, request: web.Request):
        if self.stub_master:
            return web.Response(text=self.response, status=self.http_code)
        self.msgs.put(request.query['name'])
        return web.Response(text="ok")

    async def get_annotation_info(self, request: web.Request):
        # stub annotation info
        info = {
            'start': 0,
            'end': 10e6,
            'count': 10
        }
        return web.Response(text=json.dumps(info), status=self.http_code,
                            content_type='application/json')

    async def get_annotations(self, request: web.Request):
        self.msgs.put(dict(request.query))

        resp = self.response
        if resp == "":
            resp = "[]" # so a call to get_annotations returns an empty list
        return web.Response(text=resp, status=self.http_code,
                            content_type='application/json')

    async def update_annotation(self, request: web.Request):
        if self.stub_annotation:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        title = body['title']
        content = body['content']
        # so we can check that the message was received
        self.msgs.put((title, content))
        return web.json_response(data={"stub": "value"})

    async def create_annotation(self, request: web.Request):
        if self.stub_annotation:
            return web.Response(text=self.response, status=self.http_code)
        body = await request.json()
        body["id"] = 1
        # so we can check that the message was received
        self.msgs.put((annotation.from_json(body)))
        return web.json_response(data=body)

    async def delete_annotation(self, request: web.Request):
        if self.stub_annotation:
            return web.Response(text=self.response, status=self.http_code)

        self.msgs.put(request.query['id'])
        return web.json_response(text="ok")

    async def delete_all_annotations(self, request: web.Request):
        if self.stub_annotation:
            return web.Response(text=self.response, status=self.http_code)

        self.msgs.put(dict(request.query))
        return web.Response(text="ok")


    def _find_entry(self, params):
        if 'id' in params:
            stream_id = int(params['id'])
            for entry in self.streams.values():
                if entry.stream.id == stream_id:
                    return entry
            else:
                raise ValueError
        else:
            path = params['path']
            if path not in self.streams:
                raise ValueError
            return self.streams[params['path']]


class FakeJouleTestCase(helpers.AsyncTestCase):
    def start_server(self, server):
        key = master.make_key()
        self.proc = psutil.Process()
        port = unused_port()
        self.msgs = multiprocessing.Queue()
        self.outbound_msgs = multiprocessing.Queue()
        self.server_proc = multiprocessing.Process(target=server.start,
                                                   args=(port,
                                                         self.msgs,
                                                         self.outbound_msgs,
                                                         key))
        self.server_proc.start()
        ready = False
        time.sleep(0.01)
        self.conf_dir = tempfile.TemporaryDirectory()
        while not ready:
            for conn in psutil.net_connections():
                if conn.laddr.port == port:
                    ready = True
                    break
            else:
                time.sleep(0.1)
        # create a .joule config directory with key info
        with open(os.path.join(self.conf_dir.name, ConfigFiles.nodes), 'w') as f:
            f.write(json.dumps([{"name": "fake_joule",
                                 "key": key,
                                 "url": "http://127.0.0.1:%d" % port}]))
        with open(os.path.join(self.conf_dir.name, ConfigFiles.default_node), 'w') as f:
            f.write("fake_joule")
        os.environ["JOULE_USER_CONFIG_DIR"] = self.conf_dir.name

    def stop_server(self):
        if self.server_proc is None:
            return

        self.msgs.close()
        self.msgs.join_thread()

        self.outbound_msgs.put("stop")
        self.outbound_msgs.close()
        while self.server_proc.is_alive():
            time.sleep(0.01)
        self.outbound_msgs.close()
        self.outbound_msgs.join_thread()
        # join any zombies
        self.server_proc.join()
        # force delete to close underlying pipes
        del self.outbound_msgs
        del self.msgs
        del self.server_proc
        self.conf_dir.cleanup()
        del os.environ["JOULE_USER_CONFIG_DIR"]

    def assertCliSuccess(self, cli_result):
        if cli_result.exit_code != 0:
            print("#### Error CLI Output #####: ")
            print(cli_result.output)
        self.assertEqual(cli_result.exit_code, 0)

### Helper functions
def print_result_on_error(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)


#!/usr/bin/env python3
import time
import io

import unittest
import folder, data_stream, event_stream, module, data, db, users

def main():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromModule(event_stream))
    suite.addTests(loader.loadTestsFromModule(users))
    suite.addTests(loader.loadTestsFromModule(db))
    suite.addTests(loader.loadTestsFromModule(folder))
    suite.addTests(loader.loadTestsFromModule(data_stream))
    suite.addTests(loader.loadTestsFromModule(module))
    suite.addTests(loader.loadTestsFromModule(data))
    output = io.StringIO()
    runner = unittest.TextTestRunner(failfast=True, verbosity=2)
    result = runner.run(suite)
    if result.wasSuccessful():
        print("OK")
        return 0
    else:
        print("FAIL")
        output.seek(0)
        print(output.read())
        return -1


if __name__ == '__main__':
    time.sleep(1)
    print("...running tests")
    exit(main())



import unittest
from unittest import mock
import tempfile
import shutil
import os
import asyncio

from joule.api import helpers, TcpNode
from joule import errors
from joule.constants import ConfigFiles

CAFILE = os.path.join(os.path.dirname(__file__), 'ca.joule.crt')


class TestApiHelpers(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        os.environ["JOULE_USER_CONFIG_DIR"] = self.temp_dir.name
        self.ca_path = os.path.join(self.temp_dir.name, "ca.crt")
        shutil.copyfile(CAFILE, self.ca_path)
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.temp_dir.cleanup()
        self.loop.close()

    def test_creates_config_dir(self):
        # require helper to create the directory
        config_dir = os.path.join(self.temp_dir.name, "configs")
        os.environ["JOULE_USER_CONFIG_DIR"] = config_dir
        helpers.get_nodes()
        self.assertTrue(os.path.isdir(config_dir))

    def test_normal_operation(self):
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")
        default_file = os.path.join(self.temp_dir.name, ConfigFiles.default_node)

        helpers.save_node(node1)
        helpers.save_node(node2)
        default_node = helpers.get_node()
        # first node should be the default
        self.assertEqual(node1.name, default_node.name)
        # nodes should use the cafile
        self.assertEqual(default_node.session.cafile, self.ca_path)
        # unless it is missing...
        os.remove(self.ca_path)
        retrieved_node2 = helpers.get_node("node2")
        # then nodes do not have a cafile set
        self.assertEqual(retrieved_node2.session.cafile, "")
        self.assertEqual(retrieved_node2.url, node2.url)
        # when the default node is deleted, the next one becomes the default
        helpers.delete_node(node1)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        # if the default file is missing it is automatically created
        os.remove(default_file)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        self.assertTrue(os.path.isfile(default_file))
        # but the default can be changed when the node is added back
        helpers.save_node(node1)
        node2 = helpers.get_node()
        self.assertEqual(node2.name, "node2")
        helpers.set_default_node("node1")
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node1")
        # can set the default node by object
        helpers.set_default_node(node2)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        # may not set the default node to an empty string
        with self.assertRaises(errors.ApiError) as e:
            helpers.set_default_node("")
        self.assertIn("may not be empty", str(e.exception))
        # all nodes can be retrieved
        nodes = helpers.get_nodes()
        self.assertEqual(len(nodes), 2)
        self.assertIn(nodes[0].name, ["node1", "node2"])
        self.assertIn(nodes[1].name, ["node1", "node2"])

    def test_error_conditions_on_basic_operations(self):
        # cannot use a node that does not exist
        with self.assertRaises(errors.ApiError) as e:
            helpers.get_node("does_not_exist")
        self.assertIn("does_not_exist", str(e.exception))
        self.assertIn("not available", str(e.exception))
        with self.assertRaises(errors.ApiError):
            helpers.set_default_node("does_not_exist")
        # if no nodes are available
        with self.assertRaises(errors.ApiError):
            helpers.get_node()
        # if the config file is corrupt
        with open(os.path.join(self.temp_dir.name, ConfigFiles.nodes), "w") as f:
            f.write("invalid_json")
        with self.assertRaises(errors.ApiError) as e:
            helpers.get_node()
        self.assertIn("fix syntax", str(e.exception))
        # if no node configuration location can be found
        os.environ.pop("JOULE_USER_CONFIG_DIR", None)
        os.environ.pop("HOME", None)
        os.environ.pop("APPDATA", None)
        with self.assertRaises(errors.ApiError):
            helpers.get_node()

    @mock.patch('joule.api.helpers.chown')
    def test_fixes_config_ownership(self, mock_chown):
        # when run as root, revert ownership to the user
        os.environ["SUDO_USER"]='1'
        UID = 9898
        GID = 9899
        os.environ["SUDO_UID"]=str(UID)
        os.environ["SUDO_GID"]=str(GID)
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        helpers.save_node(node1)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node1")
        # make sure all of the paths are chowned correctly
        calls = [mock.call(os.path.join(self.temp_dir.name), UID, GID),
                 mock.call(os.path.join(self.temp_dir.name, ConfigFiles.nodes), UID, GID),
                 mock.call(os.path.join(self.temp_dir.name, ConfigFiles.default_node), UID, GID)]
        self.assertEqual(len(mock_chown.call_args_list), 3)
        for call_item in mock_chown.call_args_list:
            self.assertIn(call_item, calls)

        # revert environment
        os.environ.pop("SUDO_USER")

    def test_uses_correct_config_file(self):
        # follows a preference of JOULE_USER_CONFIG_DIR, HOME, APPDATA
        # when looking for configurations
        saved_environ = os.environ.copy()
        for env_var in ["JOULE_USER_CONFIG_DIR", "HOME", "APPDATA"]:
            os.environ.pop(env_var, None)
        # check in reverse preference order
        for env_var in ["APPDATA","HOME","JOULE_USER_CONFIG_DIR"]:
            with tempfile.TemporaryDirectory() as tempdir:
                os.environ[env_var] = tempdir
                #the config directory is .joule unless it is directly specified
                if env_var in ["HOME", "APPDATA"]:
                    expected_config_dir = os.path.join(tempdir, '.joule')
                else:
                    expected_config_dir = tempdir
                self.assertEqual(helpers._user_config_dir(), expected_config_dir)
                self.assertTrue(os.path.isdir(expected_config_dir))
        # restore environment
        os.environ = saved_environ

    def test_delete_node(self):
        # nodes can be deleted by object or name
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")
        helpers.save_node(node1)
        helpers.save_node(node2)
        self.assertEqual(len(helpers.get_nodes()), 2)
        helpers.delete_node(node1)
        helpers.delete_node("node2")

        # removing a non-existent node generates an error
        with self.assertRaises(errors.ApiError):
            helpers.delete_node(node1)


    def test_save_node(self):
        pass



import click
from joule.cli.lazy_group import LazyGroup

@click.group(name="proxy",
             cls=LazyGroup,
             lazy_subcommands={
                 "list": "joule.cli.proxy.list.cli_list",
                 "info": "joule.cli.proxy.info.cli_info"
             })
def proxies():
    """Proxied site information.

    Local URL's proxied by Joule.
    """
    pass  





import asyncio
import numpy as np
import logging
from datetime import datetime, timezone
from unittest import mock
from joule.models.pipes import interval_token
from joule.models.data_stream import StreamInfo, DataStream
from joule.models.element import Element
from joule import api
from joule.errors import ConfigurationError, ApiError
from joule.client.helpers import build_network_pipes
from joule import errors
from tests import helpers
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase, MockDbEntry
from joule.errors import EmptyPipeError
log = logging.getLogger('aiohttp.access')
log.setLevel(logging.WARNING)


class TestPipeHelpers(FakeJouleTestCase):

    def test_reads_historic_data(self):
        server = FakeJoule()
        src_data = create_source_data(server)

        self.start_server(server)
        ''
        my_node = api.get_node()

        async def runner():
            pipes_in, _ = await build_network_pipes({'input': '/test/source:int16[x,y,z]'},
                                                            {},
                                                            {},
                                                            my_node,
                                                            0, 100,
                                                            force=True)
            blk1 = await pipes_in['input'].read()
            self.assertTrue(pipes_in['input'].end_of_interval)
            pipes_in['input'].consume(len(blk1))
            blk2 = await pipes_in['input'].read()
            pipes_in['input'].consume(len(blk2))
            rx_data = np.hstack((blk1,
                                 interval_token(pipes_in['input'].layout),
                                 blk2))
            np.testing.assert_array_equal(rx_data, src_data)
            with self.assertRaises(EmptyPipeError):
                await pipes_in['input'].read()
            await my_node.close()

        with self.assertLogs(level='INFO'):
            asyncio.run(runner())
        self.stop_server()
        

    def test_builds_live_input_pipes(self):
        server = FakeJoule()
        src_data = create_source_data(server, is_destination=True)

        self.start_server(server)
        loop = asyncio.new_event_loop()
        my_node = api.get_node()

        async def runner():
            pipes_in, _ = await build_network_pipes({'input': '/test/source:int16[x,y,z]'},
                                                            {},
                                                            {},
                                                            my_node,
                                                            None, None, True)
            blk1 = await pipes_in['input'].read()
            self.assertTrue(pipes_in['input'].end_of_interval)
            pipes_in['input'].consume(len(blk1))
            blk2 = await pipes_in['input'].read()
            pipes_in['input'].consume(len(blk2))
            rx_data = np.hstack((blk1,
                                 interval_token(pipes_in['input'].layout),
                                 blk2))
            np.testing.assert_array_equal(rx_data, src_data)
            with self.assertRaises(EmptyPipeError):
                await pipes_in['input'].read()
            await my_node.close()

        with self.assertLogs(level='INFO'):
            loop.run_until_complete(runner())
        self.stop_server()
        loop.close()

    def test_builds_output_pipes(self):
        server = FakeJoule()
        blk1 = helpers.create_data("int16_3", start=10, step=1, length=10)
        blk2 = helpers.create_data("int16_3", start=200, step=1, length=10)

        create_destination(server,keep_us=DataStream.KEEP_ALL)

        self.start_server(server)
        my_node = api.get_node()
        async def runner():
            _, pipes_out = await build_network_pipes({},
                                                            {'output': '/test/dest:int16[e0,e1,e2]'},
                                                            {},
                                                            my_node,
                                                            10, 210, force=True)
            await pipes_out['output'].write(blk1)
            await pipes_out['output'].close_interval()
            await pipes_out['output'].write(blk2)
            await pipes_out['output'].close_interval()
            await pipes_out['output'].close()
            await my_node.close()

        asyncio.run(runner())
        # first msg should be the data removal
        (stream_id, start, end) = self.msgs.get()
        self.assertEqual(int(stream_id), 8)
        self.assertEqual(int(start), 10)
        self.assertEqual(int(end), 210)
        # next message should be the output data
        mock_entry: MockDbEntry = self.msgs.get()
        np.testing.assert_array_equal(mock_entry.data[:len(blk1)], blk1)
        np.testing.assert_array_equal(mock_entry.data[len(blk1):], blk2)

        self.assertEqual(mock_entry.intervals, [[10, 19], [200, 209]])

        self.stop_server()

    def test_warns_before_data_removal(self):
        server = FakeJoule()
        create_destination(server)

        self.start_server(server)
        ''

        my_node = api.get_node()

        async def runner():
            with mock.patch('joule.client.helpers.pipes.click') as mock_click:
                mock_click.confirm = mock.Mock(return_value=False)
                with self.assertRaises(SystemExit):
                    await build_network_pipes({},
                                              {'output': '/test/dest:int16[e0,e1,e2]'},
                                              {},
                                              my_node,
                                              1472500708000000, 1476475108000000,
                                              force=False)
                msg = mock_click.confirm.call_args[0][0]
                # check for start time
                self.assertIn("29 Aug", msg)
                # check for end time
                self.assertIn("14 Oct", msg)

                # if end_time not specified anything *after* start is removed
                mock_click.confirm = mock.Mock(return_value=False)
                with self.assertRaises(SystemExit):
                    await build_network_pipes({},
                                              {'output': '/test/dest:int16[e0,e1,e2]'},
                                              {},
                                              my_node,
                                              1472500708000000, None,
                                              force=False)
                msg = mock_click.confirm.call_args[0][0]
                # check for start time
                self.assertIn("29 Aug", msg)
                # check for *after*
                self.assertIn("after", msg)

                # if start_time not specified anything *before* end is removed
                mock_click.confirm = mock.Mock(return_value=False)
                with self.assertRaises(SystemExit):
                    await build_network_pipes({},
                                              {'output': '/test/dest:int16[e0,e1,e2]'},
                                              {},
                                              my_node,
                                              None, 1476475108000000,
                                              force=False)
                msg = mock_click.confirm.call_args[0][0]
                # check for *before*
                self.assertIn("before", msg)
                # check for end time
                self.assertIn("14 Oct", msg)
                await my_node.close()

        # suppress "cancelled" notifications when program exits
        #with self.assertLogs(level='INFO'):
        asyncio.run(runner())
        self.stop_server()
        

    def test_creates_output_stream_if_necessary(self):
        server = FakeJoule()
        self.start_server(server)
        ''

        my_node = api.get_node()

        self.assertEqual(len(server.streams), 0)

        async def runner():
            # the destination does not exist
            with self.assertRaises(errors.ApiError):
                await my_node.data_stream_get("/test/dest")
            _, pipes_out = await build_network_pipes({},
                                                            {'output': '/test/dest:int16[e0,e1,e2]'},
                                                            {},
                                                            my_node,
                                                            None, None)
            await pipes_out['output'].close()
            # make sure the stream exists
            dest_stream = await my_node.data_stream_get("/test/dest")
            self.assertIsNotNone(dest_stream)

            await my_node.close()

        with self.assertLogs(level='INFO') as log:
            asyncio.run(runner())
        log_dump = ' '.join(log.output)
        self.assertIn('creating', log_dump)
        self.stop_server()
        

    def test_configuration_errors(self):
        server = FakeJoule()
        ''
        self.start_server(server)
        my_node = api.get_node()

        # must specify an inline configuration
        with self.assertRaises(ConfigurationError):
            asyncio.run(build_network_pipes({'input': '/test/source'},
                                                        {},
                                                        {},
                                                        my_node,
                                                        10, 20,
                                                        force=True))
        asyncio.run(my_node.close())
        self.stop_server()
        

    def test_input_errors(self):
        server = FakeJoule()
        create_source_data(server, is_destination=False)
        self.start_server(server)
        loop = asyncio.new_event_loop()
        my_node = api.get_node()

        # errors on layout differences
        with self.assertRaises(ConfigurationError) as e:
            with self.assertLogs(level='INFO'):
                loop.run_until_complete(build_network_pipes({'input': '/test/source:int16[x,y]'},
                                                            {},
                                                            {},
                                                            my_node,
                                                            10, 20, force=True))
        self.assertIn('int16_3', str(e.exception))

        # errors if live stream is requested but unavailable
        with self.assertRaises(ApiError) as e:
            loop.run_until_complete(build_network_pipes({'input': '/test/source:int16[x,y,z]'},
                                                        {},
                                                        {},
                                                        my_node, None, None))
        self.assertIn('not being produced', str(e.exception))
        loop.run_until_complete(my_node.close())
        self.stop_server()
        loop.close()

    def test_output_errors(self):
        server = FakeJoule()
        create_destination(server)
        create_source_data(server, is_destination=True)
        self.start_server(server)
        loop = asyncio.new_event_loop()
        my_node = api.get_node()

        # errors on layout differences
        with self.assertRaises(ConfigurationError) as e:
            loop.run_until_complete(build_network_pipes({},
                                                        {'output': '/test/dest:float32[x,y,z]'},
                                                        {},
                                                        my_node, None, None))
        self.assertIn('int16_3', str(e.exception))

        # errors if the stream is already being produced
        with self.assertRaises(ApiError) as e:
            loop.run_until_complete(build_network_pipes({},
                                                        {'output': '/test/source:int16[e0,e1,e2]'},
                                                        {},
                                                        my_node, None, None))
        loop.run_until_complete(my_node.close())
        self.assertIn('already being produced', str(e.exception))
        self.stop_server()
        loop.close()


def create_destination(server, keep_us=100):
    # create an empty destination stream
    dest = DataStream(id=8, name="dest", keep_us=keep_us,
                      datatype=DataStream.DATATYPE.INT16,
                      updated_at=datetime.now(timezone.utc))
    dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

    # destination has no data
    server.add_stream('/test/dest', dest, StreamInfo(None, None, 0), None)


def create_source_data(server, is_destination=False):
    # create the source stream
    src = DataStream(id=0, name="source", keep_us=100,
                     datatype=DataStream.DATATYPE.INT16,
                     is_destination=is_destination,
                      updated_at=datetime.now(timezone.utc))
    src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

    # source has 100 rows of data
    src_data = np.hstack((helpers.create_data(src.layout),
                          interval_token(src.layout),
                          helpers.create_data(src.layout)))
    src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
    server.add_stream('/test/source', src, src_info, src_data, [src_info.start, src_info.end])
    return src_data



import enum
from typing import TYPE_CHECKING, List
import numpy as np
import logging

from joule.errors import PipeError, EmptyPipeError

if TYPE_CHECKING:  
    from joule.models import (Module, DataStream)

log = logging.getLogger('joule')


class Pipe:
    """
    This encapsulates streams and connects to modules.
    Some more infos2

    .. note::

       There are many different kinds of pipes

    """

    class DIRECTION(enum.Enum):
        INPUT = enum.auto()
        OUTPUT = enum.auto()
        TWOWAY = enum.auto()

    def __init__(self, name=None, direction=None, module=None, stream=None, layout=None):
        """

        :param name: optional name for the pipe, useful for debugging
        :param direction: data flow
        :type direction: DIRECTION
        :param module: what??
        :param stream: optional stream
        :param layout: the data layout
        """
        self.name: str = name
        self.direction: Pipe.DIRECTION = direction
        self.module: 'Module' = module
        self.stream: 'DataStream' = stream
        self._layout = layout
        self.closed = False
        self.decimation_level = 1
        self.subscribers: List['Pipe'] = []
        self._failed = False

    def __repr__(self):
        return '<joule.models.Pipe name=%r direction=%r layout=%r>' % (
            self.name, self.direction.name, self.layout)

    def enable_cache(self, lines: int):
        """
        Turn on caching for pipe writes. Data is only transmitted once the cache is full.
        This improves system performance especially if :meth:`write` is called
        rapidly with short arrays. Once enabled, caching cannot be disabled.

        Args:
            lines: cache size

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot control cache on input pipes")
        raise PipeError("abstract method must be implemented by child")

    async def flush_cache(self):
        """
        Force a pipe flush even if the cache is not full. Raises an error if caching is not
        enabled.

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot control cache on input pipes")
        raise PipeError("abstract method must be implemented by child")

    async def read(self, flatten=False) -> np.ndarray:
        """
        Read stream data. By default this method returns a structured
        array with ``timestamp`` and ``data`` fields. This method is a coroutine.

        Args:
            flatten: return an unstructured array (flat 2D matrix) with timestamps in the first column

        Returns:
            numpy.ndarray

        >>> data = await pipe.read()
        [1, 2, 3]
        >>> data = await pipe.read(flatten=True)
        # the same data is returned again
        [1,2,3]
        >>> pipe.consume(len(data))
        # next call to read will return only new data
        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")

        
        raise PipeError("abstract method must be implemented by child")

    def reread_last(self):
        """
        The next read will return only unconsumed data from the previous read
        and no new data from the source. The end_of_interval flag is maintained.

        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")
       
        self._reread = True

    async def read_all(self, flatten=False, maxrows=int(1e5), error_on_overflow=False) -> np.ndarray:
        """
                Read stream data. By default this method returns a structured
                array with ``timestamp`` and ``data`` fields. The pipe is automatically closed.
                This method is a coroutine.

                Args:
                    flatten: return an unstructured array (flat 2D matrix) with timestamps in the first column
                    maxrows: the maximum number of rows to read from the pipe
                    error_on_overflow: raise a PipeError exception if pipe is not empty after reading maxrows

                Returns:
                    numpy.ndarray

                >>> data = await pipe.read_all(flatten=True)
                [1, 2, 3]
        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")

        data = None
        maxrows = int(maxrows) # make sure this is an integer
        while True:
            try:
                new_data = await self.read(flatten)
                self.consume(len(new_data))
            except PipeError:
                break
            if data is None:
                data = new_data
                if len(data) > maxrows:
                    await self.close()
                    if error_on_overflow:
                        raise PipeError("More than [%d] rows, increase maxrows or disable error_on_overflow" % maxrows)
                    return data[:maxrows]
            else:
                if len(data) + len(new_data) > maxrows:
                    await self.close()
                    if error_on_overflow:
                        raise PipeError("More than [%d] rows, increase maxrows or disable error_on_overflow" % maxrows)
                    remaining_rows = maxrows - len(data)
                    if flatten:
                        data = np.vstack((data, new_data[:remaining_rows]))
                    else:
                        data = np.hstack((data, new_data[:remaining_rows]))
                    break
                if flatten:
                    data = np.vstack((data, new_data))
                else:
                    data = np.hstack((data, new_data))
        if data is None:
            raise PipeError("No data in pipe")
        return data

    def consume(self, num_rows):
        """
        Flush data from the read buffer. The next call to :meth:`read` will
        return any unflushed data followed by new incoming data.

        Args:
            num_rows: number of rows to flush from the read buffer

        """

        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot consume from an output pipe")
        raise PipeError("abstract method must be implemented by child")

    def consume_all(self):
        """
        Flush all data from the read buffer. The next call to :meth:`read` will
        only return new incoming data.
        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot consume from an output pipe")
        raise PipeError("abstract method must be implemented by child")

    async def is_empty(self):
        """ Returns True if the pipe is empty, otherwise returns False 
           and garauntees that the next call to read will succeed, it does this by 
           calling read() and catching the EmptyPipe exception, since consume is not called
           the data will be available on the next read. By setting reread internally the next call
           to read() is essentially a no-op and provides the data retrieved by this call"""
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot check output pipes for empty condition")
        if self._reread:
            return False # data is already available
        try:
            before_flag = self.end_of_interval
            data = await self.read()
            after_flag = self.end_of_interval
            num_rows = len(data)
            #print(f"### Interval Flag {before_flag}->{after_flag}, read {num_rows} rows")
            self.reread_last() # next call to read will return this data and not try to get more

        except EmptyPipeError:
            return True
        return False

    async def not_empty(self):
        return await self.is_empty() == False
    
    def fail(self):
        self._failed = True

    @property
    def end_of_interval(self):
        return False

    async def write(self, data):
        """
        Write timestamped data to the pipe. Timestamps must be monotonically increasing
        and should not overlap with existing stream data in the database. This method is a coroutine.

        Args:
            data (numpy.ndarray): May be a structured array with ``timestamp`` and ``data`` fields
            or an unstructured array with timestamps in the first column.

        >>> await pipe.write([[1000, 2, 3],[1001, 3, 4]])

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot write to an input pipe")
        raise PipeError("abstract method must be implemented by child")

    async def close_interval(self):
        """
        Signal a break in the data stream. This should be used to indicate missing data.
        Data returned from :meth:`read` will be chunked by interval boundaries.

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot write to an input pipe")
        raise PipeError("abstract method must be implemented by child")

    def close_interval_nowait(self):
        """
        Signal a break in the data stream. This will dumped cached data and should generally
        not be used. Instead use the coroutine :meth:`close_interval`.

        """
        pass  

    def subscribe(self, pipe):
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot subscribe to an input pipe")

        self.subscribers.append(pipe)

        def unsubscribe():
            i = self.subscribers.index(pipe)
            del self.subscribers[i]

        return unsubscribe

    async def close(self):
        """
        Close the pipe. This also closes any subscribers. If ``close_cb`` is defined
        it will be executed before the subscribers are closed.

        """
        # close the pipe, optionally implemented by the child
        pass  

    def change_layout(self, layout: str):
        raise PipeError("layout cannot be changed")

    @property
    def layout(self) -> str:
        if self._layout is not None:
            return self._layout
        elif self.stream is not None:
            return self.stream.layout
        else:
            raise ValueError("pipe has no layout")

    @property
    def width(self) -> int:
        shape = self.dtype['data'].shape
        if len(shape) == 0:
            return 1
        else:
            return shape[0]

    @property
    def dtype(self) -> np.dtype:
        return compute_dtype(self.layout)

    @property
    def decimated(self) -> bool:
        if self.decimation_level > 1:
            return True
        else:
            return False

    def _apply_dtype(self, data: np.ndarray) -> np.ndarray:
        """convert [data] to the pipe's [dtype]"""
        if data.ndim == 1:
            # already a structured array just verify its data type
            if data.dtype != self.dtype:
                raise PipeError("wrong dtype for 1D (structured) array" +
                                "[%s] != req type [%s]" % (data.dtype,
                                                           self.dtype))
            return data
        elif data.ndim == 2:
            # Convert to structured array
            sarray = np.zeros(data.shape[0], dtype=self.dtype)
            try:
                sarray['timestamp'] = data[:, 0]
                # Need the squeeze in case sarray['data'] is 1 dimensional
                sarray['data'] = data[:, 1:]
                return sarray
            except (IndexError, ValueError):
                raise PipeError("wrong number of fields for this data type")
        else:
            raise PipeError("wrong number of dimensions in array")

    @staticmethod
    def _format_data(data, flatten):
        if flatten:
            return np.c_[data['timestamp'][:, None], data['data']]
        else:
            return data

    @staticmethod
    def _validate_data(data):
        if type(data) is not np.ndarray:
            raise PipeError("invalid data type must be a structured array or 2D matrix")
        # check for valid data type
        try:
            if (len(data) == 0) or len(data[0]) == 0:
                log.info("pipe write called with no data")
                return False
        except TypeError:
            raise PipeError("invalid data type must be a structured array or 2D matrix")
        return True


def interval_token(layout):
    nelem = int(layout.split('_')[1])
    token = np.zeros(1, dtype=compute_dtype(layout))
    token['timestamp'] = 0
    token['data'] = np.zeros(nelem)
    return token


def find_interval_token(raw: bytes, layout):
    token = interval_token(layout)[0]
    token_bytes = token.tobytes()
    index = raw.find(token_bytes)
    if index == -1:
        return None
    # this data *may* have an interval, need to check if the token
    # is aligned to a row
    data = np.frombuffer(raw, dtype=compute_dtype(layout))
    i = 0
    for row in data:
        if np.array_equal(row, token):
            return i * len(token_bytes), (i + 1) * len(token_bytes)
        i += 1

    # suprious match, not aligned to a row
    return None


def compute_dtype(layout: str) -> np.dtype:
    try:
        ltype = layout.split('_')[0]
        lcount = int(layout.split('_')[1])
        if ltype.startswith('int'):
            atype = '<i' + str(int(ltype[3:]) // 8)
        elif ltype.startswith('uint'):
            atype = '<u' + str(int(ltype[4:]) // 8)
        elif ltype.startswith('float'):
            atype = '<f' + str(int(ltype[5:]) // 8)
        else:
            raise ValueError()
        return np.dtype([('timestamp', '<i8'), ('data', atype, (lcount,))])
    except (IndexError, ValueError):
        raise ValueError("bad layout: [%s]" % layout)



import click

from joule.cli.lazy_group import LazyGroup

@click.group(name="module",
             cls=LazyGroup,
             lazy_subcommands={
                 "list": "joule.cli.module.list.cli_list",
                 "info": "joule.cli.module.info.cli_info",
                 "logs": "joule.cli.module.logs.cli_logs"
             })
def module():
    """Retrieve module information."""
    pass  




import warnings
import os
from click.testing import CliRunner
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main


STREAM_INFO = os.path.join(os.path.dirname(__file__), 'stream.json')
warnings.simplefilter('always')


class TestStreamRename(FakeJouleTestCase):

    def test_stream_rename(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            server.response = f.read()
        server.stub_stream_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'rename', '/folder/src', 'new name'])
        self.assertEqual(result.exit_code, 0)
        stream_attrs = self.msgs.get()
        self.assertEqual(stream_attrs['name'], "new name")
        self.stop_server()

    def test_when_stream_does_not_exist(self):
        server = FakeJoule()
        server.response = "stream does not exist"
        server.http_code = 404
        server.stub_stream_rename = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'rename', '/bad/path', 'new name'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_stream_info = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'rename', '/folder/src', 'new name'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



#!/usr/bin/env python3

import joule.client
import asyncio

import joule.errors


class Adder(joule.client.FilterModule):
    """ Add DC offset to input """
    
    def custom_args(self, parser):
        parser.add_argument("offset", type=int, default=0,
                            help="apply an offset")
        
    async def run(self, parsed_args, inputs, outputs):
        stream_in: joule = inputs["input"]
        stream_out = outputs["output"]
        while not self.stop_requested and await stream_in.not_empty():
            sarray = await stream_in.read()
            sarray["data"] += parsed_args.offset
            await asyncio.sleep(0.25)
            await stream_out.write(sarray)
            stream_in.consume(len(sarray))
            if stream_in.end_of_interval:
                print("closing interval")
                await stream_out.close_interval()
            

if __name__ == "__main__":
    r = Adder()
    r.start()



#!/usr/bin/env python3
import time
import asyncio
import numpy as np
from joule import api
from joule import errors
from joule.utilities import human_to_timestamp as h2ts
from joule.utilities import timestamp_to_human as ts2h
from click.testing import CliRunner
from joule.constants import ApiErrorMessages
import unittest
from joule.cli import main

t1 = h2ts('1 Sep 2020 00:00:00')
t2 = h2ts('2 Sep 2020 00:00:00')
t3 = h2ts('3 Sep 2020 00:00:00')
t4 = h2ts('4 Sep 2020 00:00:00')
t5 = h2ts('5 Sep 2020 00:00:00')
t6 = h2ts('6 Sep 2020 00:00:00')
t7 = h2ts('7 Sep 2020 00:00:00')
t8 = h2ts('8 Sep 2020 00:00:00')
t9 = h2ts('9 Sep 2020 00:00:00')
tA = h2ts('10 Sep 2020 00:00:00')
tB = h2ts('11 Sep 2020 00:00:00')
tC = h2ts('12 Sep 2020 00:00:00')
tD = h2ts('13 Sep 2020 00:00:00')
tE = h2ts('14 Sep 2020 00:00:00')

test_case = unittest.TestCase()

async def setup():
    node:api.BaseNode = api.get_node()
    print("loading test data...", end="")
    # put sample data into /original/stream1 and /original/stream2, both have 5 elements
    #
    # times:    1  2    3 4  5 6  7 8  9 A                      B  C  D       E
    # stream1:  |-------| |-------| |------------------------------|
    # stream2:     |---------| |-------| |----------------------|     |-------|
    # =========
    # expected:    |----| |--| |--| |--| |----------------------|   
    #              2    3 4  5 6  7 8  9 A                      B
    # Make all the time stamps 1 day apart


    # create the data streams
    stream1 = api.DataStream("stream1", elements=[api.Element(name=f"elem1_{i}") for i in range(5)])
    stream1 = await node.data_stream_create(stream1,"/original")
    pipe = await node.data_write(stream1)
    await pipe.write(_create_data(t1, t3))
    await pipe.close_interval()
    await pipe.write(_create_data(t4, t7))
    await pipe.close_interval()
    await pipe.write(_create_data(t8, tC))
    await pipe.close()

    stream2 = api.DataStream("stream2", elements=[api.Element(name=f"elem2_{i}") for i in range(5)])
    stream2 = await node.data_stream_create(stream2,"/original")
    pipe = await node.data_write(stream2)
    await pipe.write(_create_data(t2, t5))
    await pipe.close_interval()
    await pipe.write(_create_data(t6, t9))
    await pipe.close_interval()
    await pipe.write(_create_data(tA, tB))
    await pipe.close_interval()
    await pipe.write(_create_data(tD, tE))
    await pipe.close()

    # add annotations to the streams
    await node.annotation_create(api.Annotation(title="note",start=t1,end=t3,content="api annotation"), "/original/stream1")
    await node.annotation_create(api.Annotation(title="note2",start=t9,end=tA,content="api annotation"), "/original/stream2")

    # create the event streams
    events = []
    for ts in [t1, t2, t3, t4, t5, t6, t7, t8, t9, tA, tB, tC, tD, tE]:
        events.append(api.Event(start_time=ts,end_time=ts+60e6, content={"name": ts2h(ts)}))
    event_stream = await node.event_stream_create(api.EventStream(name="events",event_fields={"name": "string"}), "/original")
    await node.event_stream_write(event_stream, events)
    await node.close()
    print("OK")

def test_backup_data():
    runner = CliRunner()
    result = runner.invoke(main, "data read /original/stream1 --file /tmp/stream1.h5".split(" "))
    _assert_success(result)
    result = runner.invoke(main, "data ingest /tmp/stream1.h5 --stream /ingested/stream1".split(" "))
    _assert_success(result)
    async def _validate():
        node = api.get_node()
        orig_info = await node.data_stream_info("/original/stream1")
        new_info = await node.data_stream_info("/ingested/stream1")
        # all data should be copied over
        test_case.assertEqual(orig_info.start, new_info.start)
        test_case.assertEqual(orig_info.end, new_info.end)
        # just check to make sure the original stream has the expected data
        test_case.assertAlmostEqual(orig_info.start, t1, delta=10e6)
        test_case.assertAlmostEqual(orig_info.end, tC, delta=10e6)
        print(await node.data_intervals("/original/stream1"))
        print(await node.data_intervals("/ingested/stream1"))
        await node.close()
    asyncio.run(_validate())
    assert result.exit_code == 0

def test_process_data():
    print("running filters...")
    runner = CliRunner()
    result = runner.invoke(main, "data merge -p /original/stream1 -s /original/stream2 -d /filtered/merge1_2".split(" "), input="Y\n")
    _assert_success(result)
    result = runner.invoke(main, "data filter median -w 3 /original/stream2 /filtered/median2".split(" "), input="Y\n")
    _assert_success(result)
    result = runner.invoke(main, "data filter mean -w 3 /original/stream1 /filtered/mean1".split(" "), input="Y\n")
    _assert_success(result)

    ## run time bounded filters
    # times:    1  2    3 4  5 6  7 8  9 A                      B  C  D       E
    # stream1:  |-------| |-------| |------------------------------|
    # stream2:     |---------| |-------| |----------------------|     |-------|
    # =========
    # merge1_2:    |----| |--| |--| |--| |----------------------|        
    # median2b:           |--| |-------| |----|
    #                                         ^--A_2
    tA_2 = tA+2*60*60e6 # 2 hours after tA
    result = runner.invoke(main, f"data filter median -w 3 --start {t4} --end {tA_2} /original/stream2 /filtered/median2b".split(" "), input="Y\n")
    _assert_success(result)
  
    async def _validate():
        node = api.get_node()
        #await print_intervals(node,"/original/stream1")
        #await print_intervals(node,"/original/stream2")
        #await print_intervals(node,"/filtered/median2b")
        #await print_intervals(node,"/filtered/merge1_2")
        # should be within 10 seconds of the time bounds (depends on the sample generation by create_data)
        await assert_has_intervals(node, "/filtered/median2b", [(t4,t5),(t6,t9),(tA,tA_2)])
        await assert_has_intervals(node, "/filtered/merge1_2", [(t2,t3),(t4,t5),(t6,t7),(t8,t9),(tA,tB)])
        await node.close()
    
    asyncio.run(_validate())

def test_copy_all_data():
    print("copying all data...", end="")
    runner = CliRunner()
    result = runner.invoke(main, "folder copy /original /copy".split(" "), input="Y\n")
    _assert_success(result)
    async def _validate():
        node = api.get_node()
        orig_info = await node.data_stream_info("/original/stream1")
        copied_info = await node.data_stream_info("/copy/stream1")
        # all data should be copied over
        test_case.assertEqual(orig_info.start, copied_info.start)
        test_case.assertEqual(orig_info.end, copied_info.end)
        # annotations should be copied over
        orig_annotations = await node.annotation_get("/original/stream1")
        copied_annotations = await node.annotation_get("/copy/stream1")
        test_case.assertEqual(len(orig_annotations), len(copied_annotations))
        # event stream should be copied over
        orig_events = await node.event_stream_read_list("/original/events")
        copied_events = await node.event_stream_read_list("/copy/events")
        test_case.assertListEqual(orig_events, copied_events)
        # clean up
        await node.folder_delete("/copy", recursive=True)
        try:
            await node.folder_get("/copy")
            raise test_case.fail("folder not deleted")
        except errors.ApiError as e:
            test_case.assertIn(ApiErrorMessages.folder_does_not_exist, str(e))
        await node.close()
    asyncio.run(_validate())
    print("OK")

def test_copy_data_range():
    print("copying data range...",end="")
    runner = CliRunner()
    result = runner.invoke(main, f"folder copy /original /copy_range --start {ts2h(t8)} --end {ts2h(tB+1)}".split(" "), input="Y\n")
    _assert_success(result)
    async def _validate():
        node = api.get_node()
        copied_info = await node.data_stream_info("/copy/stream1")
        # data range should be copied over
        test_case.assertEqual(t8, copied_info.start)
        test_case.assertAlmostEqual(tB, copied_info.end, delta=2*60*1e6) # within 2 minutes
        # only annotations in the range be copied over
        copied_annotations = await node.annotation_get("/copy/stream1")
        test_case.assertEqual(len(copied_annotations), 1)
        test_case.assertEqual(copied_annotations[0].title, "note2")
        # events in the range should be copied over
        copied_events = await node.event_stream_read_list("/copy/events")
        test_case.assertListEqual([e.start_time for e in copied_events], [t8, t9, tA, tB])
        await node.close()
        # clean up
        await node.folder_delete("/copy", recursive=True)
        try:
            await node.folder_get("/copy")
            raise test_case.fail("folder not deleted")
        except errors.ApiError as e:
            test_case.assertIn(ApiErrorMessages.folder_does_not_exist, str(e))
        await node.close()
    asyncio.run(_validate())
    print("OK")

async def assert_has_intervals(node, stream, expected_intervals):
    actual_intervals = await node.data_intervals(stream)
    test_case.assertEqual(len(actual_intervals), len(expected_intervals))
    for actual, expected in zip(actual_intervals, expected_intervals):
        test_case.assertAlmostEqual(actual[0], expected[0], delta=2e6)
        test_case.assertAlmostEqual(actual[1], expected[1], delta=2e6)

async def print_intervals(node, stream):
    intervals = await node.data_intervals(stream)
    # expect /filtered/median2b to span t4 to tC
    print(f"=== {stream} intervals ====")
    for start,end in intervals:
        print(f"\t{ts2h(start)} -> {ts2h(end)}")
    print("==============================")


def _create_data(start, end):
    num_points = int((end - start)//1e6)
    step = 1e6
    """Create a random block of data with [layout] structure"""
    ts = np.arange(start, start + step * num_points, step, dtype=np.uint64)
    sarray = np.zeros(len(ts), dtype=np.dtype([('timestamp', '<i8'), ('data', '<f4', (5,))]))
    data = np.random.rand(num_points, 5)
    sarray['timestamp'] = ts
    sarray['data'] = data
    return sarray

def _assert_success(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)
    test_case.assertEqual(result.exit_code, 0)

if __name__ == "__main__":
    time.sleep(8)  # wait for jouled to boot
    asyncio.run(setup())
    test_copy_all_data()
    test_process_data()
    test_backup_data()



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp

from joule.models.supervisor import Supervisor
import joule.controllers
from tests.controllers.helpers import MockWorker
from joule import app_keys
from joule.constants import EndPoints

class TestModuleController(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        wreader = MockWorker("reader", {}, {'output': '/reader/path'}, uuid=1)
        wfilter = MockWorker("filter", {'input': '/reader/path'}, {'output': '/output/path'}, uuid=2)
        app[app_keys.supervisor] = Supervisor([wreader, wfilter], [], None)  # type: ignore
        return app


    async def test_module_list(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.modules,
                                                                 params={"statistics": 1})
        workers = await resp.json()
        self.assertEqual(len(workers), 2)
        for worker in workers:
            if worker['name'] == 'reader':
                self.assertEqual(worker['outputs']['output'], '/reader/path')
                self.assertEqual(len(worker['inputs']), 0)
            if worker['name'] == 'filter':
                self.assertEqual(worker['outputs']['output'], '/output/path')
                self.assertEqual(worker['inputs']['input'], '/reader/path')
            self.assertIn('statistics', worker)

        # can ommit statistics in the response
        resp = await self.client.request("GET", EndPoints.modules)
        workers = await resp.json()
        # check some fields in the response
        self.assertEqual(len(workers), 2)
        for worker in workers:
            if worker['name'] == 'reader':
                self.assertEqual(worker['outputs']['output'], '/reader/path')
                self.assertEqual(len(worker['inputs']), 0)
            if worker['name'] == 'filter':
                self.assertEqual(worker['outputs']['output'], '/output/path')
                self.assertEqual(worker['inputs']['input'], '/reader/path')
            self.assertNotIn('statistics', worker)


    async def test_module_info(self):
        # can retrieve info by name
        resp = await self.client.request("GET", EndPoints.module,
                                         params={'name': 'filter'})
        worker = await resp.json()
        self.assertEqual(worker['outputs']['output'], '/output/path')
        self.assertEqual(worker['inputs']['input'], '/reader/path')
        self.assertIn('statistics', worker)

        # can retrieve info by ID
        resp = await self.client.request("GET", EndPoints.module,
                                         params={'id': 2})
        worker = await resp.json()
        self.assertEqual(worker['outputs']['output'], '/output/path')
        self.assertEqual(worker['inputs']['input'], '/reader/path')
        self.assertIn('statistics', worker)


    async def test_module_logs(self):
        # can retrieve logs by name
        resp = await self.client.request("GET", EndPoints.module_logs,
                                         params={'name': 'reader'})
        logs = await resp.json()
        # mock worker produces two log entries
        self.assertEqual(len(logs), 2)

        # can retrieve logs by ID
        resp = await self.client.request("GET", EndPoints.module_logs,
                                         params={'id': 2})
        logs = await resp.json()
        # mock worker produces two log entries
        self.assertEqual(len(logs), 2)



import click
from typing import Dict, List, Tuple, Optional, Union
from operator import attrgetter
import aiohttp
import asyncio
import json
import numpy as np
import datetime

from joule.cli.config import pass_config
from joule.api import get_node, BaseNode, Annotation
from joule.api.annotation import from_json as annotation_from_json
from joule.models import data_stream, DataStream, pipes, StreamInfo
from joule.utilities import interval_difference, human_to_timestamp, timestamp_to_human
from joule import errors

Interval = Tuple[int, int]


@click.command(name="copy")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-n', '--new', help="copy starts at the last timestamp of the destination", is_flag=True)
@click.option('-d', '--destination-node', help="node name")
@click.argument("source")
@click.argument("destination")
@pass_config
def data_copy(config, start, end, new, destination_node, source, destination):
    """Copy data between two streams."""
    try:
        asyncio.run(_run(config.node, start, end, new, destination_node,
                                     source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        asyncio.run(
            config.close_node())


async def _run(config_node, start, end, new, destination_node, source, destination):
    source_node = config_node
    
    try:
        if destination_node is None:
            dest_node = config_node
        elif type(destination_node) is str:
            dest_node = get_node(destination_node)
        else:
            dest_node = destination_node
    except errors.ApiError:
        raise click.ClickException("Error: cannot connect to destination node [%s]" % destination_node)

    # retrieve the source stream
    src_stream = await _retrieve_source(source_node, source)

    # retrieve the destination stream (create it if necessary)
    dest_stream = await _retrieve_destination(dest_node, destination, src_stream)
    # make sure streams are compatible
    if src_stream.layout != dest_stream.layout:
        raise errors.ApiError("Error: source (%s) and destination (%s) datatypes are not compatible" % (
            src_stream.layout, dest_stream.layout))
    # warn if the elements are not the same
    element_warning = False
    src_elements = sorted(src_stream.elements, key=attrgetter('index'))
    dest_elements = sorted(dest_stream.elements, key=attrgetter('index'))
    for i in range(len(src_elements)):
        if src_elements[i].name != dest_elements[i].name:
            element_warning = True
        if src_elements[i].units != dest_elements[i].units:
            element_warning = True
    if (element_warning and
            not click.confirm("WARNING: Element configurations do not match. Continue?")):
        click.echo("Cancelled")
        return
    # if new is set start and end may not be specified
    if new and start is not None:
        raise click.ClickException("Error: either specify 'new' or a starting timestamp, not both")

    # make sure the time bounds make sense
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise errors.ApiError("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise errors.ApiError("invalid end time: [%s]" % end)
    if (start is not None) and (end is not None) and ((end - start) <= 0):
        raise click.ClickException("Error: start [%s] must be before end [%s]" % (
            datetime.datetime.fromtimestamp(start / 1e6),
            datetime.datetime.fromtimestamp(end / 1e6)))
    if new:
        # pull all the destination intervals and use the end of the last one as the 'end' for the copy
        dest_intervals = await dest_node.data_intervals(dest_stream, start=None, end=None)
        if len(dest_intervals) > 0:
            start = dest_intervals[-1][-1]
            print("Starting copy at [%s]" % timestamp_to_human(start))
        else:
            print("Starting copy at beginning of source")
    # compute the target intervals (source - dest)
    src_intervals = await source_node.data_intervals(src_stream, start, end)
    dest_intervals = await dest_node.data_intervals(dest_stream, start, end)
    new_intervals = interval_difference(src_intervals, dest_intervals)
    existing_intervals = interval_difference(src_intervals, new_intervals)

    async def _copy(intervals):
        # compute the duration of data to copy
        duration = 0
        for interval in intervals:
            duration += interval[1] - interval[0]

        with click.progressbar(
                label='Copying data',
                length=duration) as bar:
            for interval in intervals:
                await _copy_interval(interval[0], interval[1], bar)
                await _copy_annotations(interval[0], interval[1])

    async def _copy_annotations(istart, iend):
        src_annotations = await source_node.annotation_get(src_stream.id, start=istart, end=iend)
        
        dest_annotations = await dest_node.annotation_get(dest_stream.id, start=istart, end=iend)
        new_annotations = [a for a in src_annotations if a not in dest_annotations]
        for annotation in new_annotations:
            await dest_node.annotation_create(annotation, dest_stream.id)

    if len(new_intervals) == 0:
        if len(src_intervals) > 0:
            click.echo("Nothing to copy, syncing annotations")
            for interval in src_intervals:
                await _copy_annotations(interval[0], interval[1])
        else:
            click.echo("Nothing to copy")
        # clean up
        await dest_node.close()
        await source_node.close()
        return

    async def _copy_interval(istart, iend, bar):
        
        src_params = {'id': src_stream.id, 'start': istart, 'end': iend}
        src_url = "{server}/data".format(server=source_node.session.url)
        src_headers = {"X-API-KEY": source_node.session.key}
        src_ssl = source_node.session.ssl_context
        async with aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=None)) as session:
            async with session.get(src_url,
                                   params=src_params,
                                   headers=src_headers,
                                   ssl=src_ssl) as src_response:
                if src_response.status != 200:
                    msg = await src_response.text()
                    if msg == 'this stream has no data':
                        # This is not an error because a previous copy may have been interrupted
                        # This will cause the destination to have an interval gap where the source has no data
                        # Example:   source:  |**     *******|
                        #            dest:    |** |  |*******|
                        #                          ^--- looks like missing data but there's nothing in the source
                        return  # ignore empty intervals
                    raise click.ClickException("Error reading from source: %s" % msg)

                pipe = pipes.InputPipe(stream=dest_stream, reader=src_response.content)

                async def _data_sender():

                    last_ts = istart
                    while await pipe.not_empty():
                        data = await pipe.read()
                        pipe.consume(len(data))
                        if len(data) > 0:
                            cur_ts = data[-1]['timestamp']
                            yield data.tobytes()
                            # total time extents of this chunk
                            bar.update(cur_ts - last_ts)
                            last_ts = cur_ts
                    bar.update(iend - last_ts)

               
                dst_url = "{server}/data".format(server=dest_node.session.url)
                dst_params = {"id": dest_stream.id}
                dst_headers = {"X-API-KEY": dest_node.session.key}
                dst_ssl = dest_node.session.ssl_context
                async with session.post(dst_url,
                                        params=dst_params,
                                        data=_data_sender(),
                                        headers=dst_headers,
                                        ssl=dst_ssl,
                                        chunked=True) as dest_response:
                    if dest_response.status != 200:
                        msg = await dest_response.text()
                        raise errors.ApiError("Error writing to destination: %s" % msg)

    try:
        # copy over any new annotations from existing intervals
        for interval in existing_intervals:
            await _copy_annotations(interval[0], interval[1])
        await _copy(new_intervals)
        click.echo("\tOK")
    # this should be caught by the stream info requests
    # it is only generated if the joule server stops during the
    # data read/write
    except aiohttp.ClientError as e:  
        raise click.ClickException("Error: %s" % str(e))
    finally:
        await dest_node.close()
        await source_node.close()


async def _retrieve_source(server: BaseNode | str, path: str) -> DataStream:
    src_stream = await server.data_stream_get(path)
    src_info = await server.data_stream_info(path)

    if src_info.start is None or src_info.end is None:
        raise errors.ApiError("[%s] has no data" % path)
    return src_stream


async def _retrieve_destination(server: str | BaseNode, path: str, template: DataStream) -> DataStream:
    try:
        return await server.data_stream_get(path)
    except errors.ApiError as e:
        if "404" in str(e):
            return await _create_joule_stream(server, path, template)
        raise e


async def _create_joule_stream(node: BaseNode, path: str, template: DataStream) -> DataStream:
    click.echo("\tcreating destination stream")
    # split the destination into the path and stream name
    dest_stream = data_stream.from_json(template.to_json())
    dest_stream.keep_us = DataStream.KEEP_ALL
    dest_stream.name = path.split("/")[-1]
    folder = "/".join(path.split("/")[:-1])
    return await node.data_stream_create(dest_stream, folder)


import warnings
import dateparser
from click.testing import CliRunner
from ..fake_joule import FakeJoule, FakeJouleTestCase, print_result_on_error
from joule.cli import main
from joule.utilities import human_to_timestamp as h2ts
warnings.simplefilter('always')


class TestConsolidate(FakeJouleTestCase):

    def test_data_consolidate(self):
        server = FakeJoule()
        NUM_CONSOLIDATED = 5
        server.num_consolidated = NUM_CONSOLIDATED
        self.start_server(server)
        runner = CliRunner()
        start_str = "20 January 2015 12:00"
        end_str = "21 January 2015 12:00"
        result = runner.invoke(main, ['data', 'consolidate',
                                      '/folder/src',
                                      '--start', start_str,
                                      '--end', end_str])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        (path, start, end, max_gap) = self.msgs.get()
        self.assertEqual(path, '/folder/src')
        self.assertEqual(int(start), h2ts(start_str))
        self.assertEqual(int(end), h2ts(end_str))
        self.assertGreater(max_gap, 0) # default max_gap parameter
        # make sure decimate was not called
        self.assertEqual(self.msgs.qsize(), 0)
        # check the output for NUM_CONSOLIDATED intervals
        self.assertIn(f"{NUM_CONSOLIDATED} intervals", result.output)
        self.stop_server()

    def test_data_consolidate_and_decimate(self):
        # Run again with no intervals to consolidate, a max-gap parameter
        # and redecimate flag
        server = FakeJoule()
        server.num_consolidated = 0
        self.start_server(server)
        runner = CliRunner()
        MAX_GAP = 9191
        result = runner.invoke(main, ['data', 'consolidate',
                                      '/folder/src', '--max-gap',MAX_GAP,
                                      '--redecimate'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("No intervals less than", result.output)
        # consolidation
        (_, start, end, max_gap) = self.msgs.get()
        self.assertEqual(max_gap, 9191)
        self.assertIsNone(start)
        self.assertIsNone(end)
        # decimation
        (cmd, path) = self.msgs.get()
        self.assertEqual(cmd, "decimate")
        self.assertEqual(path, "/folder/src")
        # check the output for no intervals consolidated but redemication is run
        self.assertIn(f"No intervals less than {MAX_GAP} us", result.output)
        self.assertIn("Recomputing decimations", result.output)
        self.stop_server()

    def test_parameter_errors(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        # invalid start time
        result = runner.invoke(main, ['data', 'consolidate',
                                      '/folder/src',
                                      '--start', 'not a time'])
        self.assertEqual(result.exit_code, 1)
        self.assertIn("invalid start time", result.output)
        # invalid end time
        result = runner.invoke(main, ['data', 'consolidate',
                                      '/folder/src',
                                      '--end', 'not a time'])
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("invalid end time", result.output)
        # invalid max-gap
        result = runner.invoke(main, ['data', 'consolidate',
                                      '/folder/src',
                                      '--max-gap', 'not a number'])
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("max-gap", result.output)
        self.stop_server()

    def test_server_errors(self):
        server = FakeJoule()
        runner = CliRunner()
        server.stub_data_consolidate = True
        server.response = "server-error"
        server.http_code = 500
        self.start_server(server)

        result = runner.invoke(main, ['data', 'consolidate', '/folder/src'])
        self.assertIn("500", result.output)
        self.assertIn("server-error", result.output)
        self.stop_server()


from aiohttp.test_utils import AioHTTPTestCase
from aiohttp import web
import aiohttp
import json
import asyncio
import joule.controllers
from tests.controllers.helpers import create_db, MockSupervisor, MockWorker
from joule.models.proxy import Proxy
from joule.api.event_stream import EventStream as ApiEventStream, from_json as event_stream_from_json
from joule.api.event import Event
from joule import app_keys
from joule.constants import EndPoints
from joule.constants import ApiErrorMessages

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestAppController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([],[]))
        self.supervisor = MockSupervisor()
        self.supervisor.proxies = [Proxy('proxy1', 11, 'http://localhost:8080'),
                                   Proxy('proxy2', 22, 'http://localhost:8081')]
        app[app_keys.supervisor] = self.supervisor
        return app
    
    async def test_get_proxy(self):

        # must specify a name or id
        resp = await self.client.get(EndPoints.proxy)
        self.assertEqual(resp.status, 400)
        self.assertIn(ApiErrorMessages.specify_id_or_path, await resp.text())

        # proxy must exist
        resp = await self.client.get(EndPoints.proxy, params={'name': 'does_not_exist'})
        self.assertEqual(resp.status, 404)
        self.assertIn("proxy does not exist", await resp.text())


from joule.models.pipes import LocalPipe
import unittest
import asyncio
import numpy as np
import argparse
import pdb

import unittest
from tests import helpers
from joule.client.builtins.merge_filter import MergeFilter
from joule.errors import EmptyPipeError
from icecream import ic

WINDOW = 9
WIDTH = 4
REPS_PER_BLOCK = 9
NUM_BLOCKS = 3


class TestMergeFilter(unittest.TestCase):

    def test_offset_streams(self):
        # m :  0 1 2 3 4 5 (all 1's)
        # s1:  3 4 5       (all 2's)
        # s2:  1 2 3 4 5   (all 3's)
        # ------------------
        # out: 3 4 5       (all 1,2,3's)
        cases = [
            # CASE 1: secondarys start after primary
            [([0, 1, 2, 3, 4, 5],  # *******
              [3, 4, 5],  # ****
              [1, 2, 3, 4, 5]),  # ******
             [3, 4, 5]],  # ----
            # CASE 2: secondarys start before primary
            [([3, 4, 5],  # ****
              [2, 3, 4, 5],  # ******
              [1, 2, 3, 4, 5]),  # *******
             [3, 4, 5]],  # ----
            # CASE 3: mixed
            [([2, 3, 4, 5],  # ****
              [0, 1, 2, 3, 4, 5],  # ******
              [3, 4, 5]),  # ***
             [3, 4, 5]],  # ---
        ]

        async def run_case(inputs, expected, primary_width):
            my_filter = MergeFilter()
            primary_data = np.vstack((inputs[0], np.ones((primary_width, len(inputs[0]))))).T
            secondary1_data = np.vstack((inputs[1], 2 * np.ones((3, len(inputs[1]))))).T
            secondary2_data = np.vstack((inputs[2], 3 * np.ones((1, len(inputs[2]))))).T
            primary = LocalPipe("float32_%d" % primary_width, name="primary")
            secondary1 = LocalPipe("float32_3", name="secondary1")
            secondary2 = LocalPipe("float32_1", name="secondary2")
            output = LocalPipe("float32_%d" % (primary_width + 4), name="output")
            args = argparse.Namespace(primary="primary", pipes="unset")
            # seed the input data
            await primary.write(primary_data)
            await secondary1.write(secondary1_data)
            await secondary2.write(secondary2_data)
            await secondary1.close()
            await secondary2.close()
            await primary.close()

            # run filter in an event loop
            await my_filter.run(args,
                                inputs={'primary': primary, 'secondary1': secondary1, 'secondary2': secondary2},
                                outputs={'output': output})
            result = output.read_nowait(flatten=True)
            expected_data = np.vstack((expected,
                                       np.ones((primary_width, len(expected))),
                                       2 * np.ones((3, len(expected))),
                                       3 * np.ones((1, len(expected))))).T
            np.testing.assert_array_equal(expected_data, result)

        for case in cases:
            asyncio.run(run_case(inputs=case[0], expected=case[1], primary_width=1))
            asyncio.run(run_case(inputs=case[0], expected=case[1], primary_width=3))

    def test_early_primary(self):
        # first primary read does not overlap with any secondary data
        VISUALIZE = False

        async def _run():

            # Primary: y=10x-212
            ts = np.arange(0, 1300, 10)
            primary_data = np.array([ts, 10 * ts - 212]).T
            primary = helpers.TestingPipe("float32_1", name="primary")
            await primary.write(primary_data[:10])
            await primary.write(primary_data[10:20])
            await primary.write(primary_data[20:])
            await primary.close()

            # Secondary1: y=-3.3x+436
            ts = np.arange(500, 1000, 10)
            secondary1_data = np.array([ts, -3.3 * ts + 436]).T
            secondary1 = helpers.TestingPipe("float32_1", name="secondary1")
            await secondary1.write(secondary1_data[:20])
            await secondary1.write(secondary1_data[20:30])
            await secondary1.write(secondary1_data[30:])
            await secondary1.close()

            args = argparse.Namespace(primary="primary", pipes="unset")
            output = LocalPipe("float32_2", name="output")

            # run filter in an event loop
            my_filter = MergeFilter()

            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1},
                                outputs={'output': output})

            result = await output.read_all()
            ts = result['timestamp']
            # check that the timestamps cover the correct range
            self.assertEqual(ts[0], 500)
            self.assertEqual(ts[-1], 990)

            # check the primary
            primary_actual = result['data'][:, 0]
            residual = (10 * ts - 212) - primary_actual
            np.testing.assert_allclose(residual, 0)
            # check secondary1
            secondary1_actual = result['data'][:, 1]
            residual = (-3.3 * ts + 436) - secondary1_actual
            np.testing.assert_allclose(residual, 0)

            # NOTE: this test is close but not perfect... hmmm
            np.testing.assert_allclose(residual, 0, rtol=1e-5, atol=1e-4)

            if VISUALIZE:
                from matplotlib import pyplot as plt
                for data in [primary_data, secondary1_data]:
                    plt.plot(data[:, 0], data[:, 1], linewidth=1)

                    plt.plot(result['timestamp'], result['data'], '--', linewidth=2)
                    plt.show()

        asyncio.run(_run())

    def test_early_secondary(self):
        # first primary read does not overlap with any secondary data
        VISUALIZE = False

        async def _run():
            # Primary: y=-3x+400
            ts = np.arange(500, 1000, 10)
            primary_data = np.array([ts, -3 * ts + 400]).T
            primary = helpers.TestingPipe("float32_1", name="primary")
            # break this up so it takes multiple reads to finish
            primary.write_nowait(primary_data[:30])
            primary.write_nowait(primary_data[30:35])
            primary.write_nowait(primary_data[35:45])
            primary.write_nowait(primary_data[45:])
            await primary.close()
            # Secondary1: y=10x-212
            ts = np.arange(0, 1300, 10)
            secondary1_data = np.array([ts, 10 * ts - 212]).T
            secondary1 = helpers.TestingPipe("float32_1", name="secondary1")
            secondary1.write_nowait(secondary1_data[:10])
            secondary1.write_nowait(secondary1_data[10:20])
            secondary1.write_nowait(secondary1_data[20:])
            await secondary1.close()
            # Secondary2: y=-3.3x+436
            ts = np.arange(300, 900, 10)
            secondary2_data = np.array([ts, -3.3 * ts + 436]).T
            secondary2 = helpers.TestingPipe("float32_1", name="secondary2")
            secondary2.write_nowait(secondary2_data)
            await secondary2.close()
            args = argparse.Namespace(primary="primary", pipes="unset")
            output = helpers.TestingPipe("float32_3", name="output")

            # run filter in an event loop
            my_filter = MergeFilter()
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1,
                                        'secondary2': secondary2},
                                outputs={'output': output})
            result = output.data_blocks[0]
            ts = result['timestamp']
            # check that the timestamps cover the correct range
            self.assertEqual(ts[0], 500)
            self.assertLessEqual(ts[-1], 1000)

            # check the primary
            primary_actual = result['data'][:, 0]
            residual = (-3 * ts + 400) - primary_actual
            np.testing.assert_allclose(residual, 0)
            # check secondary1
            secondary1_actual = result['data'][:, 1]
            residual = (10 * ts - 212) - secondary1_actual
            np.testing.assert_allclose(residual, 0)

            # check secondary2
            secondary2_actual = result['data'][:, 2]
            residual = (-3.3 * ts + 436) - secondary2_actual
            # NOTE: this test is close but not perfect... hmmm
            np.testing.assert_allclose(residual, 0, rtol=1e-5, atol=1e-4)

            if VISUALIZE:
                import matplotlib
                matplotlib.use('TkAgg')
                from matplotlib import pyplot as plt

                for data in [primary_data, secondary1_data, secondary2_data]:
                    plt.plot(data[:, 0], data[:, 1], linewidth=1)

                plt.plot(result['timestamp'], result['data'], '--', linewidth=2)
                plt.show()

        asyncio.run(_run())

    def test_no_overlap(self):

        async def _run():
            # Primary: y=10x-212
            ts = np.arange(0, 500, 1)
            primary_data = np.array([ts, 10 * ts - 212]).T
            primary = helpers.TestingPipe("float32_1", name="primary")
            await primary.write(primary_data[:100])
            await primary.write(primary_data[100:200])
            await primary.write(primary_data[200:])
            await primary.close()
            # Secondary1: y=-3.3x+436
            ts = np.arange(510, 1000, 1)
            secondary1_data = np.array([ts, -5 * ts + 436]).T
            secondary1 = helpers.TestingPipe("float32_1", name="secondary1")
            await secondary1.write(secondary1_data[:150])
            await secondary1.write(secondary1_data[150:350])
            await secondary1.write(secondary1_data[350:])
            await secondary1.close()

            args = argparse.Namespace(primary="primary", pipes="unset")
            output = LocalPipe("float32_2", name="output")

            # run filter in an event loop
            my_filter = MergeFilter()
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1},
                                outputs={'output': output})

            self.assertTrue(await output.is_empty())
            with self.assertRaises(EmptyPipeError):
                await output.read()

        asyncio.run(_run())

    def test_multiple_blocks(self):
        # first primary read does not overlap with any secondary data
        VISUALIZE = False

        async def _run():
            # Primary: y=10x-212
            ts = np.arange(0, 500, 1)
            primary_data = np.array([ts, 10 * ts - 212]).T
            primary = helpers.TestingPipe("float32_1", name="primary")
            primary.write_nowait(primary_data[:100])
            primary.write_nowait(primary_data[100:200])
            primary.write_nowait(primary_data[200:])
            await primary.close()
            # Secondary1: y=-3.3x+436
            ts = np.arange(0, 500, 1)
            secondary1_data = np.array([ts, -5 * ts + 436]).T
            secondary1 = helpers.TestingPipe("float32_1", name="secondary1")
            secondary1.write_nowait(secondary1_data[:150])
            secondary1.write_nowait(secondary1_data[150:350])
            secondary1.write_nowait(secondary1_data[350:])
            await secondary1.close()

            args = argparse.Namespace(primary="primary", pipes="unset")
            output = LocalPipe("float32_2", name="output")

            # run filter in an event loop
            my_filter = MergeFilter()
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1},
                                outputs={'output': output})
            # put together the data_blocks (should not be any interval breaks)
            # remove the interval close at the end
            result = await output.read_all()
            ts = result['timestamp']
            # check that the timestamps cover the correct range
            self.assertEqual(ts[0], 0)
            self.assertEqual(ts[-1], 499)
            # no duplicate timestamps
            self.assertEqual(len(np.unique(ts)), len(primary_data))

            # check the primary
            primary_actual = result['data'][:, 0]
            residual = (10 * ts - 212) - primary_actual
            np.testing.assert_allclose(residual, 0)
            # check secondary1
            secondary1_actual = result['data'][:, 1]
            residual = (-5 * ts + 436) - secondary1_actual
            np.testing.assert_allclose(residual, 0)

            # NOTE: this test is close but not perfect... hmmm
            np.testing.assert_allclose(residual, 0, rtol=1e-5, atol=1e-4)

            if VISUALIZE:
                from matplotlib import pyplot as plt
                for data in [primary_data, secondary1_data]:
                    plt.plot(data[:, 0], data[:, 1], linewidth=1)

                plt.plot(result['timestamp'], result['data'], '--', linewidth=2)
                plt.show()

        asyncio.run(_run())

    @unittest.skip('TODO: needs more analysis')
    def test_interval_breaks(self):
        # realigns inputs after interval breaks

        # primary:[0----100] [110-----------200]   [210------300]
        # secondary1:   [50---------160] [170-----------230]
        # secondary2: [10---------------------190] [205-----------310]
        # ========================================================
        # output:   [50-100][110-160][170-190] [205-230]

        # Primary: y=10x-212
        async def _run():
            ts = np.arange(0, 500, 1)
            primary_data = np.array([ts, 10 * ts - 212]).T
            primary = LocalPipe("float32_1", name="primary")
            await primary.write(primary_data[:101])
            await primary.close_interval()
            await primary.write(primary_data[110:201])
            await primary.close_interval()
            await primary.write(primary_data[210:301])
            await primary.close()

            # Secondary1: y=-3x+436
            ts = np.arange(0, 500, 1)
            secondary_data = np.array([ts, -3 * ts + 436]).T
            secondary1 = LocalPipe("float32_1", name="secondary1")
            await secondary1.write(secondary_data[50:161])
            await secondary1.close_interval()
            await secondary1.write(secondary_data[170:231])
            await secondary1.close()

            # Secondary2: y=30x+210
            ts = np.arange(0, 500, 1)
            secondary_data = np.array([ts, 30 * ts + 210]).T
            secondary2 = LocalPipe("float32_1", name="secondary2")
            await secondary2.write(secondary_data[10:191])
            await secondary2.close_interval()
            await secondary2.write(secondary_data[205:311])
            await secondary2.close()

            output = LocalPipe("float32_3", name="output")
            my_filter = MergeFilter()
            args = argparse.Namespace(primary="primary", pipes="unset")
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1,
                                        'secondary2': secondary2},
                                outputs={'output': output})
            chunk = await output.read()
            self.assertEqual(chunk['timestamp'][0], 50)
            self.assertEqual(chunk['timestamp'][-1], 100)
            self.assertTrue(output.end_of_interval)
            output.consume(len(chunk))

            chunk = await output.read()
            self.assertEqual(chunk['timestamp'][0], 110)
            self.assertEqual(chunk['timestamp'][-1], 160)
            self.assertTrue(output.end_of_interval)
            output.consume(len(chunk))

            chunk = await output.read()
            self.assertEqual(chunk['timestamp'][0], 170)
            self.assertEqual(chunk['timestamp'][-1], 190)
            self.assertTrue(output.end_of_interval)
            output.consume(len(chunk))

            chunk = await output.read()
            self.assertEqual(chunk['timestamp'][0], 210)
            self.assertEqual(chunk['timestamp'][-1], 230)
            output.consume(len(chunk))

            self.assertTrue(output.is_empty())

        asyncio.run(_run())

    def test_different_arrival_rates(self):
        # merges multiple streams each arriving with different chunk sizes

        # All streams have the same data at the same rate, but they arrive
        # in different chunks. Expect all elements in the merged output to be the same
        VISUALIZE = False

        async def _run():
            ts = np.arange(0, 1000)
            values = np.random.randn(1000, 1)
            data = np.hstack((ts[:, None], values))

            primary = helpers.TestingPipe("float32_1", name="primary")
            secondary1 = helpers.TestingPipe("float32_1", name="secondary1")
            secondary2 = helpers.TestingPipe("float32_1", name="secondary2")
            # seed the input data
            for block in np.split(data, [200, 354, 700, 800, 930]):
                primary.write_nowait(block)
            for block in np.split(data, [155, 600, 652, 900]):
                secondary1.write_nowait(block)
            for block in np.split(data, [100, 300, 600]):
                secondary2.write_nowait(block)
            await primary.close()
            await secondary1.close()
            await secondary2.close()
            # run filter in an event loop
            my_filter = MergeFilter()
            args = argparse.Namespace(primary="primary", pipes="unset")
            output = LocalPipe("float32_3", name="output")
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1,
                                        'secondary2': secondary2},
                                outputs={'output': output})
            # put together the data_blocks (should not be any interval breaks)
            # remove the interval close at the end

            result = await output.read()
            output.consume(len(result))
            self.assertTrue(await output.is_empty())
            # all elements should match the data
            np.testing.assert_array_almost_equal(result['data'][:, 0][:, None], values)
            np.testing.assert_array_almost_equal(result['data'][:, 1][:, None], values)
            np.testing.assert_array_almost_equal(result['data'][:, 2][:, None], values)

            if VISUALIZE:
                from matplotlib import pyplot as plt
                f, (ax1, ax2) = plt.subplots(2, 1, sharey=True)
                ax1.plot(result['timestamp'], result['data'][:, 0], linewidth=4)
                ax1.plot(result['timestamp'], result['data'][:, 1], linewidth=1)
                ax1.set_title('Secondary 1 vs primary')

                ax2.plot(result['timestamp'], result['data'][:, 0], linewidth=4)
                ax2.plot(result['timestamp'], result['data'][:, 2], linewidth=1)
                ax2.set_title('Secondary 2 vs primary')

                plt.show()

        asyncio.run(_run())

    def test_static_cases(self):
        # merges streams with several different data rates
        # 1) primary == Secondary1
        # 2) primary  > Secondary2
        # 3) primary  < Secondary3
        # 4) primary != Secondary4 (timestamps randomly offset)
        VISUALIZE = False

        # primary: y=10x-212
        ts = np.arange(127, 1000, 10)
        primary_data = np.array([ts, 10 * ts - 212]).T

        # Secondary1: y=-3x+436
        secondary1_data = np.array([ts, -3 * ts + 436]).T

        # Secondary 2: y=0.5x+101
        ts = np.arange(200, 1000, 47)
        secondary2_data = np.array([ts, 0.5 * ts + 101]).T

        # Secondary 3: y=8x+3000
        ts = np.arange(0, 756, 3)
        secondary3_data = np.array([ts, 8 * ts - 3000]).T

        # Secondary 4: y=-15x+6000
        ts = np.unique([np.round(x) for x in np.random.uniform(100, 1100, 478)])
        secondary4_data = np.array([ts, -15 * ts + 6000]).T

        async def run() -> np.ndarray:
            my_filter = MergeFilter()
            primary = LocalPipe("float32_1", name="primary")
            secondary1 = LocalPipe("float32_1", name="secondary1")
            secondary2 = LocalPipe("float32_1", name="secondary2")
            secondary3 = LocalPipe("float32_1", name="secondary3")
            secondary4 = LocalPipe("float32_1", name="secondary4")
            output = LocalPipe("float32_5", name="output")
            args = argparse.Namespace(primary="primary", pipes="unset")
            # seed the input data
            primary.write_nowait(primary_data)
            secondary1.write_nowait(secondary1_data)
            secondary2.write_nowait(secondary2_data)
            secondary3.write_nowait(secondary3_data)
            secondary4.write_nowait(secondary4_data)
            [await pipe.close() for pipe in [primary, secondary1, secondary2, secondary3, secondary4]]
            # run filter in an event loop
            await my_filter.run(args,
                                inputs={'primary': primary,
                                        'secondary1': secondary1,
                                        'secondary2': secondary2,
                                        'secondary3': secondary3,
                                        'secondary4': secondary4},
                                outputs={'output': output})
            return await output.read_all()

        result = asyncio.run(run())
        ts = result['timestamp']
        # check that the timestamps cover the correct range
        self.assertGreaterEqual(ts[0], 200)
        self.assertLessEqual(ts[-1], 756)

        # check the primary
        primary_actual = result['data'][:, 0]
        residual = (10 * ts - 212) - primary_actual
        np.testing.assert_allclose(residual, 0)
        # check secondary1
        secondary1_actual = result['data'][:, 1]
        residual = (-3 * ts + 436) - secondary1_actual
        np.testing.assert_allclose(residual, 0)
        # check secondary2
        secondary2_actual = result['data'][:, 2]
        residual = (0.5 * ts + 101) - secondary2_actual
        np.testing.assert_allclose(residual, 0)
        # check secondary3
        secondary3_actual = result['data'][:, 3]
        residual = (8 * ts - 3000) - secondary3_actual
        np.testing.assert_allclose(residual, 0)

        if VISUALIZE:
            from matplotlib import pyplot as plt
            for data in [primary_data, secondary1_data,
                         secondary2_data, secondary3_data, secondary4_data]:
                plt.plot(data[:, 0], data[:, 1], linewidth=1)

            plt.plot(result['timestamp'], result['data'], '--', linewidth=2)
            plt.show()



from sqlalchemy import create_engine
from sqlalchemy.orm import Session
import unittest
import logging
import configparser
from unittest import mock

from tests.helpers import DbTestCase
from joule.models import Base
from joule.services import load_streams

logger = logging.getLogger('joule')

class TestLoadStreamErrors(DbTestCase):

    @mock.patch('joule.services.load_streams.load_configs')
    def test_validates_path(self, load_configs: mock.Mock):
        bad_paths = ["", "/slash/at/end/", "bad name", "/double/end//",
                     "//double/start", "/*bad&symb()ls"]
        for path in bad_paths:
            parser = configparser.ConfigParser()
            parser.read_string("""
                                [Main]
                                name=Bad DataStream
                                DataType=float32
                                Path = %s
                                [Element1]
                                name=x
                                """ % path)
            load_configs.return_value = {'stream.conf': parser}
            with self.assertLogs(level="ERROR"):
                load_streams.run('', self.db)

        good_paths = ["/", "/short", "/meters-4/prep-a",
                      "/meter_4/prep-b", "/path  with/ spaces"]
        for path in good_paths:
            parser = configparser.ConfigParser()
            parser.read_string("""
                               [Main]
                                name=Bad DataStream
                                DataType=float32
                                Path = %s
                                [Element1]
                                name=x
                               """ % path)
            load_configs.return_value = {'stream.conf': parser}
            new_streams = load_streams.run('', self.db)
            self.assertEqual(len(new_streams), 1)

    @mock.patch('joule.services.load_streams.load_configs')
    def test_logs_invalid_streams(self, load_configs):
            parser = configparser.ConfigParser()
            parser.read_string("""
                        [Main]
                        name=Bad DataStream
                        DataType=float32
                        Path = /no/elements
                        """)
            load_configs.return_value = {'stream.conf': parser}
            with self.assertLogs(level="ERROR"):
                new_streams = load_streams.run('', self.db)
                self.assertEqual(len(new_streams), 0)

    @mock.patch('joule.services.load_streams.load_configs')
    def test_checks_path_is_present(self, load_configs):
        parser = configparser.ConfigParser()
        parser.read_string("""
                            [Main]
                            name=No Path Setting
                            DataType=float32
                            [Element1]
                            name = x
                            """)
        load_configs.return_value = {'stream.conf': parser}
        with self.assertLogs(level="ERROR"):
            new_streams = load_streams.run('', self.db)
            self.assertEqual(len(new_streams), 0)



from typing import Dict
from aiohttp import web
from aiohttp.test_utils import unused_port
import numpy as np
import multiprocessing
from joule.constants import EndPoints
from tests import helpers


class FakeStream:
    def __init__(self, layout, start=None, end=None, rows=0, intervals=None):
        self.layout = layout
        self.start = start
        self.end = end
        self.rows = rows
        self.intervals = intervals
        self.metadata = {}
        if start is not None and end is not None:
            if self.intervals is None:
                self.intervals = [[start, end]]

    @property
    def dtype(self):
        ltype = self.layout.split('_')[0]
        lcount = int(self.layout.split('_')[1])
        atype = '??'
        if ltype.startswith('int'):
            atype = '<i' + str(int(ltype[3:]) // 8)
        elif ltype.startswith('uint'):
            atype = '<u' + str(int(ltype[4:]) // 8)
        elif ltype.startswith('float'):
            atype = '<f' + str(int(ltype[5:]) // 8)
        return np.dtype([('timestamp', '<i8'), ('data', atype, lcount)])



import subprocess
import shlex
import socket
import time

def run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL):
    return subprocess.run(shlex.split(cmd), stdout=stdout, stderr=stderr)

def wait_for_joule_host(host, port=80, timeout=5):
    i=0
    while i < timeout:
        i+=1
        try:
            s = socket.create_connection((host, port))
            s.close()
            return
        except ConnectionRefusedError:
            time.sleep(1)

    raise Exception(f"timeout waiting for {host}:{port}")


from joule.models.pipes import InputPipe, OutputPipe, LocalPipe
from joule.models.pipes.factories import reader_factory, writer_factory
from joule.errors import PipeError
import os
import numpy as np
import asyncio
import unittest.mock
from tests import helpers

"""
Tests NumpyPipes with subprocess pipes 
"""


class TestStreamingPipes(helpers.AsyncTestCase):

    def test_pipes_numpy_arrays(self):
        LAYOUT = "int16_2"
        LENGTH = 1000
        (fd_r, fd_w) = os.pipe()
        ''
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r))
        npipe_out = OutputPipe(layout=LAYOUT, writer_factory=writer_factory(fd_w))
        test_data = helpers.create_data(LAYOUT, length=LENGTH)

        async def writer():
            for block in helpers.to_chunks(test_data, 270):
                await npipe_out.write(block)
                await asyncio.sleep(0.01)

        async def reader():
            blk_size = 357
            data_cursor = 0  # index into test_data
            run_count = 0
            while data_cursor != len(test_data):
                # consume all data in pipe
                data = await npipe_in.read()
                rows_used = min(len(data), blk_size)
                used_data = data[:rows_used]
                # print(used_data['data'][:,1])
                npipe_in.consume(rows_used)
                start_pos = data_cursor
                end_pos = data_cursor + len(used_data)
                np.testing.assert_array_equal(used_data,
                                              test_data[start_pos:end_pos])

                data_cursor += len(used_data)
                run_count += 1

            # make sure the reads are broken up (otherwise test is lame...)
            self.assertGreater(run_count, 2)

        async def runner():
            await asyncio.gather(writer(),reader())
            await npipe_in.close()
            await npipe_out.close()

        asyncio.run(runner())

    def test_reconstructs_fragmented_data(self):
        LAYOUT = "int16_2"
        LENGTH = 1000
        (fd_r, fd_w) = os.pipe()
        ''
        # wrap the file descriptors in numpypipes
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r))
        test_data = helpers.create_data(LAYOUT, length=LENGTH)

        async def writer():
            b_data = test_data.tobytes()
            f = open(fd_w, 'wb', 0)
            f.write(b_data[:395])
            await asyncio.sleep(0.05)
            f.write(b_data[395:713])
            await asyncio.sleep(0.05)
            f.write(b_data[713:])
            f.close()

        async def reader():
            data_cursor = 0  # index into test_data
            run_count = 0
            while data_cursor != len(test_data):
                # consume all data in pipe
                data = await npipe_in.read()
                npipe_in.consume(len(data))
                start_pos = data_cursor
                end_pos = data_cursor + len(data)
                np.testing.assert_array_equal(data,
                                              test_data[start_pos:end_pos])
                data_cursor += len(data)
                run_count += 1

            # make sure the reads are broken up (otherwise test is lame...)
            self.assertGreater(run_count, 2)

        async def runner():
            await asyncio.gather(writer(), reader())
            await npipe_in.close()

        asyncio.run(runner())

    def test_read_data_must_be_consumed(self):
        """writes to pipe sends data to reader and any subscribers"""
        LAYOUT = "float32_2"
        LENGTH = 500
        UNCONSUMED_ROWS = 4
        BUFFER_SIZE = 100
        (fd_r, fd_w) = os.pipe()
        # wrap the file descriptors in numpypipes
        ''
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r),
                             buffer_size=BUFFER_SIZE)
        npipe_out = OutputPipe(layout=LAYOUT, writer_factory=writer_factory(fd_w))

        test_data = helpers.create_data(LAYOUT, length=LENGTH)

        async def writer():
            await npipe_out.write(test_data)

        async def reader():
            data = await npipe_in.read()
            self.assertLessEqual(len(data), BUFFER_SIZE)
            repeat = await npipe_in.read()
            np.testing.assert_array_equal(data, repeat)
            npipe_in.consume(len(data) - UNCONSUMED_ROWS)

            next_data = await npipe_in.read()

            np.testing.assert_array_equal(data[-UNCONSUMED_ROWS:],
                                          next_data[:UNCONSUMED_ROWS])

        async def runner():
            await asyncio.gather(writer(), reader())
            await npipe_in.close()
            await npipe_out.close()

        asyncio.run(runner())

    def test_raises_consume_errors(self):
        LAYOUT = "float32_2"
        LENGTH = 100
        (fd_r, fd_w) = os.pipe()
        ''
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r))
        npipe_out = OutputPipe(layout=LAYOUT, writer_factory=writer_factory(fd_w))
        test_data = helpers.create_data(LAYOUT, length=LENGTH)

        # cannot consume more data than is in the pipe

        async def runner():
            await npipe_out.write(test_data)
            read_data = await npipe_in.read()
            await npipe_in.close()
            await npipe_out.close()
            return read_data
        read_data = asyncio.run(runner())

        # can't consume more than was read
        with self.assertRaises(PipeError) as e:
            npipe_in.consume(len(read_data)+1)
        self.assertTrue('consume' in str(e.exception))
        # can't consume less than zero
        with self.assertRaises(PipeError) as e:
            npipe_in.consume(-1)
        # fine to consume zero rows
        npipe_in.consume(0)
        self.assertTrue('negative' in str(e.exception))

    def test_caching(self):
        LAYOUT = "float32_2"
        LENGTH = 128
        CACHE_SIZE = 40
        (fd_r, fd_w) = os.pipe()
        ''
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r))
        npipe_in.TIMEOUT_INTERVAL = 2
        npipe_out = OutputPipe(layout=LAYOUT, writer_factory=writer_factory(fd_w))
        npipe_out.enable_cache(CACHE_SIZE)
        test_data = helpers.create_data(LAYOUT, length=LENGTH, start=1, step=1)

        async def writer():
            for block in helpers.to_chunks(test_data, 4):
                await asyncio.sleep(.1)
                await npipe_out.write(block)
            await npipe_out.flush_cache()
            # closing the interval should flush the data
            await npipe_out.close_interval()
            # add a dummy section after the interval break
            await npipe_out.write(np.ones((35, 3)))
            # closing the pipe should flush the cache
            await npipe_out.close()

        num_reads = 0

        async def reader():
            nonlocal num_reads
            index = 0
            while True:
                data = await npipe_in.read()
                npipe_in.consume(len(data))
                # make sure the data is correct
                np.testing.assert_array_equal(test_data[index:index+len(data)],
                                              data)
                num_reads += 1
                index += len(data)

                if index == len(test_data):
                    break
            # now get the dummy section after the interval break
            data = await npipe_in.read(flatten=True)
            np.testing.assert_array_equal(data, np.ones((35, 3)))

        async def runner():
            await asyncio.gather(reader(), writer())
            await npipe_in.close()
            await npipe_out.close()
        asyncio.run(runner())
        self.assertLessEqual(num_reads, np.ceil(LENGTH/CACHE_SIZE))


    def test_sends_data_to_subscribers(self):
        LAYOUT = "float32_2"
        (fd_r, fd_w) = os.pipe()
        ''
        output_cb = unittest.mock.AsyncMock()
        input_cb = unittest.mock.AsyncMock()
        subscriber_cb = unittest.mock.AsyncMock()
        npipe_out = OutputPipe(layout=LAYOUT, writer_factory=writer_factory(fd_w),
                               close_cb=output_cb)
        subscriber = LocalPipe(layout=LAYOUT, close_cb=subscriber_cb)
        npipe_out.subscribe(subscriber)
        test_data = helpers.create_data(LAYOUT)
        npipe_in = InputPipe(layout=LAYOUT, reader_factory=reader_factory(fd_r),
                             close_cb=input_cb)
        async def runner():
            await npipe_out.write(test_data)
            rx_data = await npipe_in.read()
            await npipe_in.close()
            await npipe_out.close()
            return rx_data
        rx_data = asyncio.run(runner())
        # data should be available on the InputPipe side

        np.testing.assert_array_equal(test_data, rx_data)
        # data should also be available on the Subscriber output
        rx_data = subscriber.read_nowait()
        np.testing.assert_array_equal(test_data, rx_data)

        # subscriber should be closed
        self.assertTrue(subscriber.closed)
        # make sure all of the callbacks have been executed
        self.assertEqual(output_cb.call_count, 1)
        self.assertEqual(input_cb.call_count, 1)
        self.assertEqual(subscriber_cb.call_count, 1)

    def test_invalid_write_inputs(self):
        LAYOUT = "int16_2"
        ''
        my_pipe = OutputPipe(layout=LAYOUT)

        async def test():
            with self.assertLogs(level="INFO"):
                await my_pipe.write(np.array([[]]))
            with self.assertRaises(PipeError):
                await my_pipe.write(np.array([1, 2, 3]))

        asyncio.run(test())



import click
from joule.cli.lazy_group import LazyGroup

@click.group(name="master",
             cls=LazyGroup,
             lazy_subcommands={"list": "joule.cli.master.list.cli_list",
                               "delete": "joule.cli.master.delete.cli_delete",
                               "add": "joule.cli.master.add.cli_add"})
def master():
    """Manage node access."""
    pass  




import_stmts = []
with open("base_imports.txt") as f:
    f.readline() # ignore the first line
    for line in f:
        fields =  line.split('|')
        import_stmts.append(fields[-1])

# create python click command to read in a file 
# and compare each line to the import_stmts list
# if the line ends with an item in this list then
# ignore it, otherwise write it out to a file that
# ends with _diff.txt
import click
@click.command()
@click.argument('file', type=click.File('r'))
def diff_imports(file):
    with open(file.name.replace('.txt', f'_diff.txt'), 'w') as f:
        for line in file:
            if not any(line.endswith(stmt) for stmt in import_stmts):
                f.write(line)

if __name__ == '__main__':
    diff_imports()


import unittest
import tempfile
import os
import stat

from joule.services.helpers import load_configs


class TestHelpers(unittest.TestCase):

    def test_parses_conf_files_in_path(self):
        """parses files ending in *.conf and ignores others"""
        file_names = ['stream1.conf', 'ignored',
                      'temp.conf~', 'streamA-3.conf']

        with tempfile.TemporaryDirectory() as conf_dir:
            for name in file_names:
                # create a stub stream configuration (needed for
                # configparser)
                with open(os.path.join(conf_dir, name), 'w') as f:
                    f.write('[Main]\n')

            configs = load_configs(conf_dir)
        self.assertEqual(2, len(configs))
        self.assertIn('stream1.conf', configs.keys())
        self.assertIn('streamA-3.conf', configs.keys())

    def test_logs_invalid_files(self):
        """parses files ending in *.conf and ignores others"""

        with tempfile.TemporaryDirectory() as conf_dir:
            with open(os.path.join(conf_dir, 'setting.conf'), 'w') as f:
                f.write('invalid\n')
            with open(os.path.join(conf_dir, 'noread.conf'), 'w') as f:
                f.write('[Main]\n')
            os.chmod(os.path.join(conf_dir, 'noread.conf'), stat.S_IWUSR)
            with self.assertLogs(level="ERROR"):
                configs = load_configs(conf_dir)
                self.assertEqual(len(configs), 0)


import unittest
import json
from joule import api, errors
from joule.utilities import human_to_timestamp as h2ts
from joule.utilities import timestamp_to_human as ts2h

t1 = h2ts('1 Sep 2020 00:00:00')
t2 = h2ts('2 Sep 2020 00:00:00')
t3 = h2ts('3 Sep 2020 00:00:00')
t4 = h2ts('4 Sep 2020 00:00:00')
t5 = h2ts('5 Sep 2020 00:00:00')
t6 = h2ts('6 Sep 2020 00:00:00')
t7 = h2ts('7 Sep 2020 00:00:00')
t8 = h2ts('8 Sep 2020 00:00:00')
t9 = h2ts('9 Sep 2020 00:00:00')
tA = h2ts('10 Sep 2020 00:00:00')
tB = h2ts('11 Sep 2020 00:00:00')
tC = h2ts('12 Sep 2020 00:00:00')
tD = h2ts('13 Sep 2020 00:00:00')
tE = h2ts('14 Sep 2020 00:00:00')
tF = h2ts('15 Sep 2020 00:00:00')
TS_LIST = [t1, t2, t3, t4, t5, t6, t7, t8, t9, tA, tB, tC, tD, tE, tF]
class TestStreamMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()
        """
        └── original
            └── events
        """
        events = []
        for ts in TS_LIST:
            events.append(api.Event(start_time=ts,end_time=ts+60e6, content={"name": ts2h(ts), "value": ts}))
        event_stream = await self.node.event_stream_create(api.EventStream(name="events",event_fields={"name": "string","value": "numeric"}), "/original")
        await self.node.event_stream_write(event_stream, events)
        await self.node.close()

    async def asyncTearDown(self):
        await self.node.folder_delete("/original", recursive=True)
        await self.node.close()

    async def test_stream_move(self):
        await self.node.event_stream_move("/original/events", "/test/deeper")
        info = await self.node.event_stream_info("/test/deeper/events")
        self.assertEqual(info.event_count,15)

    async def test_stream_read(self):
        from joule.api import event_stream
        event_stream.EVENT_READ_BLOCK_SIZE = 4 # force multiple backend reads

        ## can read one at a time
        i = 0
        async for event in self.node.event_stream_read("/original/events"):
            self.assertEqual(event.start_time, TS_LIST[i])
            i += 1
        self.assertEqual(i, 15)

        ## can read in blocks
        i = 0
        async for events in self.node.event_stream_read("/original/events", block_size=5):
            self.assertEqual(len(events), 5)
            i+=5
        self.assertEqual(i, 15)

        ## can read as list
        events = await self.node.event_stream_read_list("/original/events")
        self.assertEqual(len(events), 15)

        ## too many events generates a warning
        with self.assertLogs(level='WARN'):
            events = await self.node.event_stream_read_list("/original/events", limit=10)
        self.assertEqual(len(events), 10)

        ## can read by time range
        events = await self.node.event_stream_read_list("/original/events", start=t5+1, end=tD+1, include_on_going_events=False)
        self.assertEqual(len(events), 8)
        self.assertEqual(events[0].start_time, t6)
        self.assertEqual(events[-1].start_time, tD)

        ## include on going events
        events = await self.node.event_stream_read_list("/original/events", start=t5+1, end=tD+1, include_on_going_events=True)
        self.assertEqual(len(events), 9)
        self.assertEqual(events[0].start_time, t5)
        self.assertEqual(events[-1].start_time, tD)
        
        ## can filter by content
        events = await self.node.event_stream_read_list("/original/events", json_filter=json.dumps([[['value','lte',t2]]]))
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].start_time, t1)
        self.assertEqual(events[-1].start_time, t2)

    async def test_event_delete(self):
        # can remove an interior time range
        await self.node.event_stream_remove("/original/events", start=t5, end=tD+1)
        expected_events = [t1, t2, t3, t4, tE, tF]
        events = await self.node.event_stream_read_list("/original/events")
        self.assertListEqual([e.start_time for e in events], expected_events)

        # can remove the beginning
        await self.node.event_stream_remove("/original/events", end=t2+1)
        expected_events = [t3, t4, tE, tF]
        events = await self.node.event_stream_read_list("/original/events")
        self.assertListEqual([e.start_time for e in events], expected_events)

        # can remove the end: NOTE: delete includes ongoing events
        await self.node.event_stream_remove("/original/events", start=tE+1)
        expected_events = [t3, t4]
        events = await self.node.event_stream_read_list("/original/events")
        self.assertListEqual([e.start_time for e in events], expected_events)

        # can remove by JSON filter
        await self.node.event_stream_remove("/original/events", json_filter=json.dumps([[['value','lt',t4]]]))
        expected_events = [t4]
        events = await self.node.event_stream_read_list("/original/events")
        self.assertListEqual([e.start_time for e in events], expected_events)

        # can remove all the events
        await self.node.event_stream_remove("/original/events")
        events = await self.node.event_stream_read_list("/original/events")
        self.assertEqual(len(events), 0)

        # can remove the stream itself
        await self.node.event_stream_delete("/original/events")
        with self.assertRaises(errors.ApiError):
            await self.node.event_stream_info("/original/events")

        


import click
from joule.cli.lazy_group import LazyGroup

@click.group(name="data",
             cls=LazyGroup,
             lazy_subcommands={"copy": "joule.cli.data.copy.data_copy",
                               "read": "joule.cli.data.read.cmd",
                               "delete": "joule.cli.data.delete.data_delete",
                               "intervals": "joule.cli.data.intervals.intervals",
                               "consolidate": "joule.cli.data.consolidate.consolidate",
                               "merge": "joule.cli.data.merge.merge",
                               "ingest": "joule.cli.data.ingest.ingest"})
def data():
    """Interact with data streams."""
    pass  

@click.group(name="filter",
             cls=LazyGroup,
             lazy_subcommands={"mean": "joule.cli.data.mean.mean",
                               "median": "joule.cli.data.median.median"})
def filter():
    """Filter stream data."""
    pass  

data.add_command(filter)


import unittest
import datetime
from joule.models import (EventStream, Folder, event_stream)
from joule.models.data_store.event_store import StreamInfo
from joule.errors import ConfigurationError
from tests import helpers


class TestStream(unittest.TestCase):

    def test_attribute_errors(self):
        my_stream = EventStream(name='test',
                                updated_at=datetime.datetime.now())
        for bad_val in ['abc', -2, [], {}, (1, 2, 3)]:
            with self.assertRaisesRegex(ConfigurationError, 'keep'):
                my_stream.update_attributes({"keep_us": bad_val})

        for bad_val in ['abc', -1, [], {}, (1, 2, 3)]:
            with self.assertRaisesRegex(ConfigurationError, 'chunk_duration') as error:
                my_stream.update_attributes({"chunk_duration_us": bad_val})

        for bad_val in ['abc', -1, [], (1, 2, 3), {'field1': 'incorrect'}]:
            with self.assertRaisesRegex(ConfigurationError, 'event_fields') as error:
                my_stream.update_attributes({"event_fields": bad_val})

        for bad_val in ['/invalid', '', 'also/invalid']:
            with self.assertRaisesRegex(ConfigurationError, 'name') as error:
                my_stream.update_attributes({"name": bad_val})



"""
Configuration data structure for joule
Use load_configs to retrieve Configs object
"""

from joule.models import Proxy
from typing import Optional, List
from dataclasses import dataclass

DEFAULT_CONFIG = {
    "Main":
        {
            "Name": "joule_node",
            "ModuleDirectory": "/etc/joule/module_configs",
            "StreamDirectory": "/etc/joule/stream_configs",
            "SocketDirectory": "/tmp/joule",
            #"Database": "joule@localhost:5432/joule",
            "InsertPeriod": 5,
            "CleanupPeriod": 60,
            "MaxLogLines": 100,
            "EchoModuleLogs": "false"
        },
    "Proxies": {}
}


class SecurityConfig:
    def __init__(self,
                 certfile: str,
                 keyfile: str,
                 cafile: str):
        self.certfile = certfile
        self.keyfile = keyfile
        self.cafile = cafile

@dataclass
class JouleConfig:
    name: str
    module_directory: str
    stream_directory: str
    ip_address: Optional[str]
    port: Optional[int]
    socket_directory: str
    database: str
    insert_period: int
    cleanup_period: int
    max_log_lines: int
    users_file: Optional[str]
    proxies: List['Proxy']
    security: Optional['SecurityConfig']
    echo_module_logs: bool
